# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import IntEnum
from typing import Dict, Generic, Optional, TypeVar

from grpc import ChannelCredentials, ssl_channel_credentials
from ovmsclient.tfs_compat.base.serving_client import ServingClient

from datumaro.components.errors import DatumaroError, MediaTypeError
from datumaro.components.launcher import LauncherWithModelInterpreter
from datumaro.components.media import Image


class InferenceServerType(IntEnum):
    """Types of the dedicated inference server"""

    ovms = 0
    triton = 1


class ProtocolType(IntEnum):
    """Protocol type for communication with dedicated inference server"""

    grpc = 0
    http = 1


@dataclass(frozen=True)
class TLSConfig:
    """TLS configuration dataclass

    Parameters:
        client_key_path: Path to client key file
        client_cert_path: Path to client certificate file
        server_cert_path: Path to server certificate file
    """

    client_key_path: str
    client_cert_path: str
    server_cert_path: str

    def as_dict(self) -> Dict[str, str]:
        return {
            "client_key_path": self.client_key_path,
            "client_cert_path": self.client_cert_path,
            "server_cert_path": self.server_cert_path,
        }

    def as_grpc_creds(self) -> ChannelCredentials:
        server_cert, client_cert, client_key = ServingClient._prepare_certs(
            self.server_cert_path, self.client_cert_path, self.client_key_path
        )
        return ssl_channel_credentials(
            root_certificates=server_cert, private_key=client_key, certificate_chain=client_cert
        )


TClient = TypeVar("TClient")


class LauncherForDedicatedInferenceServer(Generic[TClient], LauncherWithModelInterpreter):
    """Inference launcher for dedicated inference server

    Parameters:
        model_name: Name of the model. It should match with the model name loaded in the server instance.
        model_interpreter_path: Python source code path which implements a model interpreter.
            The model interpreter implement pre-processing of the model input and post-processing of the model output.
        model_version: Version of the model loaded in the server instance
        host: Host address of the server instance
        port: Port number of the server instance
        timeout: Timeout limit during communication between the client and the server instance
        tls_config: Configuration required if the server instance is in the secure mode
        protocol_type: Communication protocol type with the server instance
    """

    def __init__(
        self,
        model_name: str,
        model_interpreter_path: str,
        model_version: int = 0,
        host: str = "localhost",
        port: int = 9000,
        timeout: float = 10.0,
        tls_config: Optional[TLSConfig] = None,
        protocol_type: ProtocolType = ProtocolType.grpc,
    ):
        super().__init__(model_interpreter_path=model_interpreter_path)

        self.model_name = model_name
        self.model_version = model_version
        self.url = f"{host}:{port}"
        self.timeout = timeout
        self.tls_config = tls_config
        self.protocol_type = protocol_type

        try:
            self._client = self._init_client()
            self._check_server_health()
            self._init_metadata()
        except Exception as e:
            raise DatumaroError(
                f"Health check failed for model_name={self.model_name}, "
                f"model_version={self.model_version}, url={self.url} and tls_config={self.tls_config}"
            ) from e

    def _init_client(self) -> TClient:
        raise NotImplementedError()

    def _check_server_health(self) -> None:
        raise NotImplementedError()

    def _init_metadata(self) -> None:
        raise NotImplementedError()

    def type_check(self, item):
        if not isinstance(item.media, Image):
            raise MediaTypeError(f"Media type should be Image, Current type={type(item.media)}")
        return True



# Copyright (C) 2019-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

# pylint: disable=unused-variable

import cv2
import numpy as np

from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image
from datumaro.util import take_by

__all__ = ["RISE"]


class RISE:
    """
    Implements RISE: Randomized Input Sampling for
    Explanation of Black-box Models algorithm.
    See explanations at: https://arxiv.org/pdf/1806.07421.pdf
    """

    def __init__(
        self,
        model,
        num_masks: int = 100,
        mask_size: int = 7,
        prob: float = 0.5,
        batch_size: int = 1,
    ):
        assert prob >= 0 and prob <= 1
        self.model = model
        self.num_masks = num_masks
        self.mask_size = mask_size
        self.prob = prob
        self.batch_size = batch_size

    def normalize_saliency(self, saliency):
        normalized_saliency = np.empty_like(saliency)
        for idx, sal in enumerate(saliency):
            normalized_saliency[idx, ...] = (sal - np.min(sal)) / (np.max(sal) - np.min(sal))
        return normalized_saliency

    def generate_masks(self, image_size):
        cell_size = np.ceil(np.array(image_size) / self.mask_size).astype(np.int8)
        up_size = tuple([(self.mask_size + 1) * cs for cs in cell_size])

        grid = np.random.rand(self.num_masks, self.mask_size, self.mask_size) < self.prob
        grid = grid.astype("float32")

        masks = np.empty((self.num_masks, *image_size))
        for i in range(self.num_masks):
            # Random shifts
            x = np.random.randint(0, cell_size[0])
            y = np.random.randint(0, cell_size[1])

            # Linear upsampling and cropping
            masks[i, ...] = cv2.resize(grid[i], up_size, interpolation=cv2.INTER_LINEAR)[
                x : x + image_size[0], y : y + image_size[1]
            ]

        return masks

    def generate_masked_dataset(self, image, image_size, masks):
        input_image = cv2.resize(image, image_size, interpolation=cv2.INTER_LINEAR)

        items = []
        for id, mask in enumerate(masks):
            masked_image = np.expand_dims(mask, axis=-1) * input_image
            items.append(
                DatasetItem(
                    id=id,
                    media=Image.from_numpy(masked_image),
                )
            )
        return Dataset.from_iterable(items)

    def apply(self, image, progressive=False):
        assert len(image.shape) in [2, 3], "Expected an input image in (H, W, C) format"
        if len(image.shape) == 3:
            assert image.shape[2] in [3, 4], "Expected BGR or BGRA input"
        image = image[:, :, :3].astype(np.float32)

        model = self.model

        image_size = model.inputs[0].shape
        logit_size = model.outputs[0].shape

        batch_size = image_size[0]
        if image_size[1] in [1, 3]:  # for CxHxW
            image_size = (image_size[2], image_size[3])
        elif image_size[3] in [1, 3]:  # for HxWxC
            image_size = (image_size[1], image_size[2])

        masks = self.generate_masks(image_size=image_size)
        masked_dataset = self.generate_masked_dataset(image, image_size, masks)

        saliency = np.zeros((logit_size[1], *image_size), dtype=np.float32)
        for batch_id, batch in enumerate(take_by(masked_dataset, batch_size)):
            outputs = model.launch(batch)

            for sample_id in range(len(batch)):
                mask = masks[batch_size * batch_id + sample_id]
                for class_idx in range(logit_size[1]):
                    score = outputs[sample_id][class_idx].attributes["score"]
                    saliency[class_idx, ...] += score * mask

                # [TODO] wonjuleee: support DRISE for detection model explainability
                # if isinstance(self.target, Label):
                #     logits = outputs[sample_id][0].vector
                #     max_score = logits[self.target.label]
                # elif isinstance(self.target, Bbox):
                #     preds = outputs[sample_id][0]
                #     max_score = 0
                #     for box in preds:
                #         if box[0] == self.target.label:
                #             confidence, box = box[1], box[2]
                #             score = iou(self.target.get_bbox, box) * confidence
                #             if score > max_score:
                #                 max_score = score
                # saliency += max_score * mask

                if progressive:
                    yield self.normalize_saliency(saliency)

        yield self.normalize_saliency(saliency)



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
from pathlib import Path
from typing import List, Optional

from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import IMAGE_EXTENSIONS, find_images


class ImageDirImporter(Importer):
    """
    Reads images from a directory as a dataset.
    """

    DETECT_CONFIDENCE = FormatDetectionConfidence.EXTREME_LOW

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--subset",
            help="The name of the subset for the produced dataset items " "(default: none)",
        )
        return parser

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        path = Path(context.root_path)
        for item in path.iterdir():
            if item.is_dir():
                context.fail("Only flat image directories are supported")
            elif item.suffix.lower() not in IMAGE_EXTENSIONS:
                context.fail(f"File {item} is not an image.")
        return super().detect(context)

    @classmethod
    def find_sources(cls, path):
        path = Path(path)
        if not path.is_dir():
            return []

        return [{"url": str(path), "format": ImageDirBase.NAME}]

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return list(IMAGE_EXTENSIONS)


class ImageDirBase(SubsetBase):
    def __init__(
        self,
        url: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(subset=subset, ctx=ctx)
        url = Path(url)
        assert url.is_dir(), url

        for path in find_images(str(url)):
            item_id = Path(path).stem
            self._items.append(
                DatasetItem(id=item_id, subset=self._subset, media=Image.from_file(path=path))
            )
        self._ann_types = set()

    @property
    def is_stream(self) -> bool:
        return True


class ImageDirExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"

    def _apply_impl(self):
        os.makedirs(self._save_dir, exist_ok=True)

        for item in self._extractor:
            if item.media:
                self._save_image(item)
            else:
                log.debug("Item '%s' has no image info", item.id)



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os.path as osp
from glob import glob
from typing import List, Optional

from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME
from datumaro.components.errors import DatasetNotFoundError
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.merge.extractor_merger import ExtractorMerger
from datumaro.plugins.data_formats.coco.base import (
    CocoCaptionsBase,
    CocoImageInfoBase,
    CocoInstancesBase,
    CocoLabelsBase,
    CocoPanopticBase,
    CocoPersonKeypointsBase,
    CocoStuffBase,
)
from datumaro.plugins.data_formats.coco.extractor_merger import COCOExtractorMerger

from .format import CocoImporterType, CocoTask


class CocoImporter(Importer):
    _TASKS = {
        CocoTask.instances: CocoInstancesBase,
        CocoTask.person_keypoints: CocoPersonKeypointsBase,
        CocoTask.captions: CocoCaptionsBase,
        CocoTask.labels: CocoLabelsBase,
        CocoTask.image_info: CocoImageInfoBase,
        CocoTask.panoptic: CocoPanopticBase,
        CocoTask.stuff: CocoStuffBase,
    }
    _IMPORTER_TYPE = CocoImporterType.default
    _ANNO_EXT = ".json"

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--keep-original-category-ids",
            action="store_true",
            help="Add dummy label categories so that category indices "
            "correspond to the category IDs in the original annotation "
            "file",
        )
        return parser

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> FormatDetectionConfidence:
        num_tasks = 0
        for task in cls._TASKS.keys():
            try:
                context.require_files(f"annotations/{task.name}_*{cls._ANNO_EXT}")
                num_tasks += 1
            except Exception:
                pass
        if num_tasks > 1:
            log.warning(
                "Multiple COCO tasks are detected. The detected format will be `coco` instead."
            )
            return FormatDetectionConfidence.MEDIUM
        else:
            context.raise_unsupported()

    def __call__(self, path, stream: bool = False, **extra_params):
        subsets = self.find_sources(path)

        if len(subsets) == 0:
            raise DatasetNotFoundError(path, self.NAME)

        # TODO: should be removed when proper label merging is implemented
        conflicting_types = {
            CocoTask.instances,
            CocoTask.person_keypoints,
            CocoTask.labels,
            CocoTask.panoptic,
            CocoTask.stuff,
        }
        ann_types = set(t for s in subsets.values() for t in s) & conflicting_types
        if 1 <= len(ann_types):
            selected_ann_type = sorted(ann_types, key=lambda x: x.name)[0]
        if 1 < len(ann_types):
            log.warning(
                "Not implemented: "
                "Found potentially conflicting source types with labels: %s. "
                "Only one type will be used: %s"
                % (", ".join(t.name for t in ann_types), selected_ann_type.name)
            )

        sources = []
        for subset, ann_files in subsets.items():
            for ann_type, ann_file in ann_files.items():
                if ann_type in conflicting_types:
                    if ann_type is not selected_ann_type:
                        log.warning(
                            "Not implemented: " "conflicting source '%s' is skipped." % ann_file
                        )
                        continue
                log.info("Found a dataset at '%s'" % ann_file)

                options = dict(extra_params)
                options["coco_importer_type"] = self._IMPORTER_TYPE
                options["subset"] = subset

                if stream:
                    options["stream"] = True

                sources.append(
                    {
                        "url": ann_file,
                        "format": self._TASKS[ann_type].NAME,
                        "options": options,
                    }
                )

        return sources

    @classmethod
    def find_sources(cls, path):
        def detect_coco_task(filename):
            for task in CocoTask:
                if filename.startswith(task.name + "_"):
                    return task
            return None

        if osp.isfile(path):
            if len(cls._TASKS) == 1:
                return {"": {next(iter(cls._TASKS)): path}}

            subset_paths = [path] if path.endswith(".json") else []
        else:
            subset_paths = glob(osp.join(path, "**", "*_*.json"), recursive=True)

        subsets = {}
        for subset_path in subset_paths:
            ann_type = detect_coco_task(osp.basename(subset_path))

            if ann_type not in cls._TASKS:
                log.warning(
                    "File '%s' was skipped, could't match this file "
                    "with any of these tasks: %s"
                    % (subset_path, ",".join(e.NAME for e in cls._TASKS.values()))
                )
                continue

            parts = osp.splitext(osp.basename(subset_path))[0].split(
                ann_type.name + "_", maxsplit=1
            )
            subset_name = parts[1] if len(parts) == 2 else DEFAULT_SUBSET_NAME
            subsets.setdefault(subset_name, {})[ann_type] = subset_path

        return subsets

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]

    @property
    def can_stream(self) -> bool:
        return True

    def get_extractor_merger(self) -> Optional[ExtractorMerger]:
        return COCOExtractorMerger


class CocoImageInfoImporter(CocoImporter):
    _TASK = CocoTask.image_info
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> FormatDetectionConfidence:
        context.require_file(f"annotations/{cls._TASK.name}_*{cls._ANNO_EXT}")
        return FormatDetectionConfidence.LOW


class CocoCaptionsImporter(CocoImageInfoImporter):
    _TASK = CocoTask.captions
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}


class CocoInstancesImporter(CocoImageInfoImporter):
    _TASK = CocoTask.instances
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}


class CocoPersonKeypointsImporter(CocoImageInfoImporter):
    _TASK = CocoTask.person_keypoints
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}


class CocoLabelsImporter(CocoImageInfoImporter):
    _TASK = CocoTask.labels
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}


class CocoPanopticImporter(CocoImageInfoImporter):
    _TASK = CocoTask.panoptic
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}


class CocoStuffImporter(CocoImageInfoImporter):
    _TASK = CocoTask.stuff
    _TASKS = {_TASK: CocoImporter._TASKS[_TASK]}



# Copyright (C) 2021-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from copy import deepcopy

import numpy as np

from datumaro.components.annotation import (
    AnnotationType,
    GroupType,
    LabelCategories,
    TabularCategories,
)
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.errors import (
    AttributeDefinedButNotFound,
    BrokenAnnotation,
    EmptyCaption,
    EmptyLabel,
    FarFromAttrMean,
    FarFromCaptionMean,
    FarFromLabelMean,
    FewSamplesInAttribute,
    FewSamplesInCaption,
    FewSamplesInLabel,
    ImbalancedAttribute,
    ImbalancedCaptions,
    ImbalancedDistInAttribute,
    ImbalancedDistInCaption,
    ImbalancedDistInLabel,
    ImbalancedLabels,
    InvalidValue,
    LabelDefinedButNotFound,
    MissingAnnotation,
    MissingAttribute,
    MissingLabelCategories,
    MultiLabelAnnotations,
    NegativeLength,
    OnlyOneAttributeValue,
    OnlyOneLabel,
    OutlierInCaption,
    RedundanciesInCaption,
    UndefinedAttribute,
    UndefinedLabel,
)
from datumaro.components.validator import Severity, TaskType, Validator
from datumaro.util import parse_str_enum_value
from datumaro.util.tabular_util import emoji_pattern

DEFAULT_LABEL_GROUP = "default"


class _TaskValidator(Validator, CliPlugin):
    DEFAULT_FEW_SAMPLES_THR = 1
    DEFAULT_IMBALANCE_RATIO_THR = 50
    DEFAULT_FAR_FROM_MEAN_THR = 5
    DEFAULT_DOMINANCE_RATIO_THR = 0.8
    DEFAULT_TOPK_BINS = 0.1

    # statistics templates
    numerical_stat_template = {
        "items_far_from_mean": {},
        "mean": None,
        "stdev": None,
        "min": None,
        "max": None,
        "median": None,
        "histogram": {
            "bins": [],
            "counts": [],
        },
        "distribution": [],
    }

    """
    A base class for task-specific validators.

    Attributes
    ----------
    task_type : str or TaskType
        task type (ie. classification, detection, segmentation, tabular)
    """

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-fs",
            "--few-samples-thr",
            default=cls.DEFAULT_FEW_SAMPLES_THR,
            type=int,
            help="Threshold for giving a warning for minimum number of "
            "samples per class (default: %(default)s)",
        )
        parser.add_argument(
            "-ir",
            "--imbalance-ratio-thr",
            default=cls.DEFAULT_IMBALANCE_RATIO_THR,
            type=int,
            help="Threshold for giving data imbalance warning. "
            "IR(imbalance ratio) = majority/minority "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-m",
            "--far-from-mean-thr",
            default=cls.DEFAULT_FAR_FROM_MEAN_THR,
            type=float,
            help="Threshold for giving a warning that data is far from mean. "
            "A constant used to define mean +/- k * standard deviation "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-dr",
            "--dominance-ratio-thr",
            default=cls.DEFAULT_DOMINANCE_RATIO_THR,
            type=float,
            help="Threshold for giving a warning for bounding box imbalance. "
            "Dominace_ratio = ratio of Top-k bin to total in histogram "
            "(default: %(default)s)",
        )
        parser.add_argument(
            "-k",
            "--topk-bins",
            default=cls.DEFAULT_TOPK_BINS,
            type=float,
            help="Ratio of bins with the highest number of data "
            "to total bins in the histogram. A value in the range [0, 1] "
            "(default: %(default)s)",
        )
        return parser

    def __init__(
        self,
        task_type,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        """
        Validator

        Parameters
        ---------------
        few_samples_thr: int
            minimum number of samples per class
            warn user when samples per class is less than threshold
        imbalance_ratio_thr: int
            ratio of majority attribute to minority attribute
            warn user when annotations are unevenly distributed
        far_from_mean_thr: float
            constant used to define mean +/- m * stddev
            warn user when there are too big or small values
        dominance_ratio_thr: float
            ratio of Top-k bin to total
            warn user when dominance ratio is over threshold
        topk_bins: float
            ratio of selected bins with most item number to total bins
            warn user when values are not evenly distributed
        """
        self.task_type = parse_str_enum_value(task_type, TaskType, default=TaskType.classification)

        if self.task_type == TaskType.classification:
            self.ann_types = {AnnotationType.label}
            self.str_ann_type = "label"
        elif self.task_type == TaskType.detection:
            self.ann_types = {AnnotationType.bbox}
            self.str_ann_type = "bounding box"
        elif self.task_type == TaskType.segmentation:
            self.ann_types = {AnnotationType.mask, AnnotationType.polygon, AnnotationType.ellipse}
            self.str_ann_type = "mask or polygon or ellipse"
        elif self.task_type == TaskType.tabular:
            self.ann_types = {AnnotationType.label, AnnotationType.caption}
            self.str_ann_type = "label or caption"

        if few_samples_thr is None:
            few_samples_thr = self.DEFAULT_FEW_SAMPLES_THR

        if imbalance_ratio_thr is None:
            imbalance_ratio_thr = self.DEFAULT_IMBALANCE_RATIO_THR

        if far_from_mean_thr is None:
            far_from_mean_thr = self.DEFAULT_FAR_FROM_MEAN_THR

        if dominance_ratio_thr is None:
            dominance_ratio_thr = self.DEFAULT_DOMINANCE_RATIO_THR

        if topk_bins is None:
            topk_bins = self.DEFAULT_TOPK_BINS

        self.few_samples_thr = few_samples_thr
        self.imbalance_ratio_thr = imbalance_ratio_thr
        self.far_from_mean_thr = far_from_mean_thr
        self.dominance_thr = dominance_ratio_thr
        self.topk_bins_ratio = topk_bins

    def _compute_common_statistics(self, dataset):
        defined_attr_template = {"items_missing_attribute": [], "distribution": {}}
        undefined_attr_template = {"items_with_undefined_attr": [], "distribution": {}}
        undefined_label_template = {
            "count": 0,
            "items_with_undefined_label": [],
        }

        stats = {
            "label_distribution": {
                "defined_labels": {},
                "undefined_labels": {},
            },
            "attribute_distribution": {"defined_attributes": {}, "undefined_attributes": {}},
        }
        stats["total_ann_count"] = 0
        stats["items_missing_annotation"] = []

        label_dist = stats["label_distribution"]
        defined_label_dist = label_dist["defined_labels"]
        undefined_label_dist = label_dist["undefined_labels"]

        attr_dist = stats["attribute_distribution"]
        defined_attr_dist = attr_dist["defined_attributes"]
        undefined_attr_dist = attr_dist["undefined_attributes"]

        label_categories = dataset.categories().get(AnnotationType.label, LabelCategories())
        base_valid_attrs = label_categories.attributes

        for category in label_categories:
            defined_label_dist[category.name] = 0

        filtered_anns = []
        for item in dataset:
            item_key = (item.id, item.subset)
            annotations = []
            for ann in item.annotations:
                if ann.type in self.ann_types:
                    annotations.append(ann)
            ann_count = len(annotations)
            filtered_anns.append((item_key, annotations))

            if ann_count == 0:
                stats["items_missing_annotation"].append(item_key)
            stats["total_ann_count"] += ann_count

            for ann in annotations:
                if not 0 <= ann.label < len(label_categories):
                    label_name = ann.label

                    label_stats = undefined_label_dist.setdefault(
                        ann.label, deepcopy(undefined_label_template)
                    )
                    label_stats["items_with_undefined_label"].append(item_key)

                    label_stats["count"] += 1
                    valid_attrs = set()
                    missing_attrs = set()
                else:
                    label_name = label_categories[ann.label].name
                    defined_label_dist[label_name] += 1

                    defined_attr_stats = defined_attr_dist.setdefault(label_name, {})

                    valid_attrs = base_valid_attrs.union(label_categories[ann.label].attributes)
                    ann_attrs = getattr(ann, "attributes", {}).keys()
                    missing_attrs = valid_attrs.difference(ann_attrs)

                    for attr in valid_attrs:
                        defined_attr_stats.setdefault(attr, deepcopy(defined_attr_template))

                for attr in missing_attrs:
                    attr_dets = defined_attr_stats[attr]
                    attr_dets["items_missing_attribute"].append(item_key)

                for attr, value in ann.attributes.items():
                    if attr not in valid_attrs:
                        undefined_attr_stats = undefined_attr_dist.setdefault(label_name, {})
                        attr_dets = undefined_attr_stats.setdefault(
                            attr, deepcopy(undefined_attr_template)
                        )
                        attr_dets["items_with_undefined_attr"].append(item_key)
                    else:
                        attr_dets = defined_attr_stats[attr]

                    attr_dets["distribution"].setdefault(str(value), 0)
                    attr_dets["distribution"][str(value)] += 1

        return stats, filtered_anns

    def _generate_common_reports(self, stats):
        """
        Validates the dataset for classification tasks based on its statistics.

        Parameters
        ----------
        dataset : IDataset object
        stats: Dict object

        Returns
        -------
        reports (list): List of validation reports (DatasetValidationError).
        """

        reports = []

        # report for dataset
        reports += self._check_missing_label_categories(stats)

        # report for item
        reports += self._check_missing_annotation(stats)

        # report for label
        reports += self._check_undefined_label(stats)
        reports += self._check_label_defined_but_not_found(stats)
        reports += self._check_only_one_label(stats)
        reports += self._check_few_samples_in_label(stats)
        reports += self._check_imbalanced_labels(stats)

        # report for attributes
        attr_dist = stats["attribute_distribution"]
        defined_attr_dist = attr_dist["defined_attributes"]
        undefined_attr_dist = attr_dist["undefined_attributes"]

        defined_labels = defined_attr_dist.keys()
        for label_name in defined_labels:
            attr_stats = defined_attr_dist[label_name]

            reports += self._check_attribute_defined_but_not_found(label_name, attr_stats)

            for attr_name, attr_dets in attr_stats.items():
                reports += self._check_missing_attribute(label_name, attr_name, attr_dets)
                reports += self._check_only_one_attribute(label_name, attr_name, attr_dets)
                reports += self._check_few_samples_in_attribute(label_name, attr_name, attr_dets)
                reports += self._check_imbalanced_attribute(label_name, attr_name, attr_dets)

        for label_name, attr_stats in undefined_attr_dist.items():
            for attr_name, attr_dets in attr_stats.items():
                reports += self._check_undefined_attribute(label_name, attr_name, attr_dets)

        return reports

    def _generate_validation_report(self, error, *args, **kwargs):
        return [error(*args, **kwargs)]

    def _check_missing_label_categories(self, stats):
        validation_reports = []

        if len(stats["label_distribution"]["defined_labels"]) == 0:
            validation_reports += self._generate_validation_report(
                MissingLabelCategories, Severity.error
            )

        return validation_reports

    def _check_missing_annotation(self, stats):
        validation_reports = []

        items_missing = stats["items_missing_annotation"]
        for item_id, item_subset in items_missing:
            validation_reports += self._generate_validation_report(
                MissingAnnotation, Severity.warning, item_id, item_subset, self.str_ann_type
            )

        return validation_reports

    def _check_missing_attribute(self, label_name, attr_name, attr_dets):
        validation_reports = []

        items_missing_attr = attr_dets["items_missing_attribute"]
        for item_id, item_subset in items_missing_attr:
            details = (item_subset, label_name, attr_name)
            validation_reports += self._generate_validation_report(
                MissingAttribute, Severity.warning, item_id, *details
            )

        return validation_reports

    def _check_undefined_label(self, stats):
        validation_reports = []

        undefined_label_dist = stats["label_distribution"]["undefined_labels"]
        for label_name, label_stats in undefined_label_dist.items():
            for item_id, item_subset in label_stats["items_with_undefined_label"]:
                details = (item_subset, label_name)
                validation_reports += self._generate_validation_report(
                    UndefinedLabel, Severity.error, item_id, *details
                )

        return validation_reports

    def _check_undefined_attribute(self, label_name, attr_name, attr_dets):
        validation_reports = []

        items_with_undefined_attr = attr_dets["items_with_undefined_attr"]
        for item_id, item_subset in items_with_undefined_attr:
            details = (item_subset, label_name, attr_name)
            validation_reports += self._generate_validation_report(
                UndefinedAttribute, Severity.error, item_id, *details
            )

        return validation_reports

    def _check_label_defined_but_not_found(self, stats):
        validation_reports = []
        count_by_defined_labels = stats["label_distribution"]["defined_labels"]
        labels_not_found = [
            label_name for label_name, count in count_by_defined_labels.items() if count == 0
        ]

        for label_name in labels_not_found:
            validation_reports += self._generate_validation_report(
                LabelDefinedButNotFound, Severity.warning, label_name
            )

        return validation_reports

    def _check_attribute_defined_but_not_found(self, label_name, attr_stats):
        validation_reports = []
        attrs_not_found = [
            attr_name
            for attr_name, attr_dets in attr_stats.items()
            if len(attr_dets["distribution"]) == 0
        ]

        for attr_name in attrs_not_found:
            details = (label_name, attr_name)
            validation_reports += self._generate_validation_report(
                AttributeDefinedButNotFound, Severity.warning, *details
            )

        return validation_reports

    def _check_only_one_label(self, stats):
        validation_reports = []
        count_by_defined_labels = stats["label_distribution"]["defined_labels"]
        labels_found = [
            label_name for label_name, count in count_by_defined_labels.items() if count > 0
        ]

        if len(labels_found) == 1:
            validation_reports += self._generate_validation_report(
                OnlyOneLabel, Severity.info, labels_found[0]
            )

        return validation_reports

    def _check_only_one_attribute(self, label_name, attr_name, attr_dets):
        validation_reports = []
        values = list(attr_dets["distribution"].keys())

        if len(values) == 1:
            details = (label_name, attr_name, values[0])
            validation_reports += self._generate_validation_report(
                OnlyOneAttributeValue, Severity.info, *details
            )

        return validation_reports

    def _check_few_samples_in_label(self, stats):
        validation_reports = []
        thr = self.few_samples_thr

        defined_label_dist = stats["label_distribution"]["defined_labels"]
        labels_with_few_samples = [
            (label_name, count)
            for label_name, count in defined_label_dist.items()
            if 0 < count <= thr
        ]

        for label_name, count in labels_with_few_samples:
            validation_reports += self._generate_validation_report(
                FewSamplesInLabel, Severity.info, label_name, count
            )

        return validation_reports

    def _check_few_samples_in_attribute(self, label_name, attr_name, attr_dets):
        validation_reports = []
        thr = self.few_samples_thr

        attr_values_with_few_samples = [
            (attr_value, count)
            for attr_value, count in attr_dets["distribution"].items()
            if count <= thr
        ]

        for attr_value, count in attr_values_with_few_samples:
            details = (label_name, attr_name, attr_value, count)
            validation_reports += self._generate_validation_report(
                FewSamplesInAttribute, Severity.info, *details
            )

        return validation_reports

    def _check_imbalanced_labels(self, stats):
        validation_reports = []
        thr = self.imbalance_ratio_thr

        defined_label_dist = stats["label_distribution"]["defined_labels"]
        count_by_defined_labels = [count for label, count in defined_label_dist.items()]

        if len(count_by_defined_labels) == 0:
            return validation_reports

        count_max = np.max(count_by_defined_labels)
        count_min = np.min(count_by_defined_labels)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= thr:
            validation_reports += self._generate_validation_report(ImbalancedLabels, Severity.info)

        return validation_reports

    def _check_imbalanced_attribute(self, label_name, attr_name, attr_dets):
        validation_reports = []
        thr = self.imbalance_ratio_thr

        count_by_defined_attr = list(attr_dets["distribution"].values())
        if len(count_by_defined_attr) == 0:
            return validation_reports

        count_max = np.max(count_by_defined_attr)
        count_min = np.min(count_by_defined_attr)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= thr:
            validation_reports += self._generate_validation_report(
                ImbalancedAttribute, Severity.info, label_name, attr_name
            )

        return validation_reports


class ClassificationValidator(_TaskValidator):
    """
    A specific validator class for classification task.
    """

    def __init__(
        self,
        task_type=TaskType.classification,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        super().__init__(
            task_type=task_type,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_ratio_thr=dominance_ratio_thr,
            topk_bins=topk_bins,
        )

    def compute_statistics(self, dataset):
        """
        Computes statistics of the dataset for the classification task.

        Parameters
        ----------
        dataset : IDataset object

        Returns
        -------
        stats (dict): A dict object containing statistics of the dataset.
        """

        stats, filtered_anns = self._compute_common_statistics(dataset)

        label_cat = dataset.categories()[AnnotationType.label]
        label_groups = label_cat.label_groups

        label_name_to_group = {}
        for label_group in label_groups:
            for idx, label_name in enumerate(label_group.labels):
                if label_group.group_type == GroupType.EXCLUSIVE:
                    label_name_to_group[label_name] = label_group.name
                else:
                    label_name_to_group[label_name] = label_group.name + f"_{idx}"

        undefined_label_name = list(stats["label_distribution"]["undefined_labels"].keys())

        stats["items_with_multiple_labels"] = []
        for item_key, anns in filtered_anns:
            occupied_groups = set()
            for ann in anns:
                if ann.label in undefined_label_name:
                    continue
                label_name = label_cat[ann.label].name
                label_group = label_name_to_group.get(label_name, DEFAULT_LABEL_GROUP)
                if label_group in occupied_groups:
                    stats["items_with_multiple_labels"].append(item_key)
                    break
                occupied_groups.add(label_group)

        return stats

    def generate_reports(self, stats):
        """
        Validates the dataset for classification tasks based on its statistics.

        Parameters
        ----------
        dataset : IDataset object
        stats: Dict object

        Returns
        -------
        reports (list): List of validation reports (DatasetValidationError).
        """

        reports = self._generate_common_reports(stats)
        reports += self._check_multi_label_annotations(stats)

        return reports

    def _check_multi_label_annotations(self, stats):
        validation_reports = []

        items_with_multiple_labels = stats["items_with_multiple_labels"]
        for item_id, item_subset in items_with_multiple_labels:
            validation_reports += self._generate_validation_report(
                MultiLabelAnnotations, Severity.error, item_id, item_subset
            )

        return validation_reports


class DetectionValidator(_TaskValidator):
    """
    A specific validator class for detection task.
    """

    def __init__(
        self,
        task_type=TaskType.detection,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        super().__init__(
            task_type=task_type,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_ratio_thr=dominance_ratio_thr,
            topk_bins=topk_bins,
        )

        self.point_template = {
            "width": deepcopy(self.numerical_stat_template),
            "height": deepcopy(self.numerical_stat_template),
            "area(wxh)": deepcopy(self.numerical_stat_template),
            "ratio(w/h)": deepcopy(self.numerical_stat_template),
            "short": deepcopy(self.numerical_stat_template),
            "long": deepcopy(self.numerical_stat_template),
        }

    def compute_statistics(self, dataset):
        """
        Computes statistics of the dataset for the detection task.

        Parameters
        ----------
        dataset : IDataset object

        Returns
        -------
        stats (dict): A dict object containing statistics of the dataset.
        """

        stats, filtered_items = self._compute_common_statistics(dataset)

        stats["items_with_negative_length"] = {}
        stats["items_with_invalid_value"] = {}
        stats["point_distribution_in_label"] = {}
        stats["point_distribution_in_attribute"] = {}
        stats["point_distribution_in_dataset_item"] = {}

        self.items = filtered_items

        def _generate_ann_bbox_info(_x, _y, _w, _h, area, ratio, _short, _long):
            return {
                "x": _x,
                "y": _y,
                "width": _w,
                "height": _h,
                "area(wxh)": area,
                "ratio(w/h)": ratio,
                "short": _short,
                "long": _long,
            }

        def _update_bbox_stats_by_label(item_key, ann, bbox_label_stats):
            bbox_has_error = False

            _x, _y, _w, _h = ann.get_bbox()
            area = ann.get_area()

            if _h != 0 and _h != float("inf"):
                ratio = _w / _h
            else:
                ratio = float("nan")

            _short = _w if _w < _h else _h
            _long = _w if _w > _h else _h

            ann_bbox_info = _generate_ann_bbox_info(_x, _y, _w, _h, area, ratio, _short, _long)

            items_w_invalid_val = stats["items_with_invalid_value"]
            for prop, val in ann_bbox_info.items():
                if val == float("inf") or np.isnan(val):
                    bbox_has_error = True
                    anns_w_invalid_val = items_w_invalid_val.setdefault(item_key, {})
                    invalid_props = anns_w_invalid_val.setdefault(ann.id, [])
                    invalid_props.append(prop)

            items_w_neg_len = stats["items_with_negative_length"]
            for prop in ["width", "height"]:
                val = ann_bbox_info[prop]
                if val < 1:
                    bbox_has_error = True
                    anns_w_neg_len = items_w_neg_len.setdefault(item_key, {})
                    neg_props = anns_w_neg_len.setdefault(ann.id, {})
                    neg_props[prop] = val

            if not bbox_has_error:
                ann_bbox_info.pop("x")
                ann_bbox_info.pop("y")
                self._update_prop_distributions(ann_bbox_info, bbox_label_stats)

            return ann_bbox_info, bbox_has_error

        # Collect property distribution
        label_categories = dataset.categories().get(AnnotationType.label, LabelCategories())
        self._compute_prop_dist(label_categories, stats, _update_bbox_stats_by_label)

        # Compute property statistics from distribution
        dist_by_label = stats["point_distribution_in_label"]
        dist_by_attr = stats["point_distribution_in_attribute"]
        self._compute_prop_stats_from_dist(dist_by_label, dist_by_attr)

        def _is_valid_bbox(item_key, ann):
            has_defined_label = 0 <= ann.label < len(label_categories)
            if not has_defined_label:
                return False

            bbox_has_neg_len = ann.id in stats["items_with_negative_length"].get(item_key, {})
            bbox_has_invalid_val = ann.id in stats["items_with_invalid_value"].get(item_key, {})
            return not (bbox_has_neg_len or bbox_has_invalid_val)

        def _update_bbox_props_far_from_mean(item_key, ann):
            base_valid_attrs = label_categories.attributes
            valid_attrs = base_valid_attrs.union(label_categories[ann.label].attributes)
            label_name = label_categories[ann.label].name
            bbox_label_stats = dist_by_label[label_name]

            _x, _y, _w, _h = ann.get_bbox()
            area = ann.get_area()
            ratio = _w / _h
            _short = _w if _w < _h else _h
            _long = _w if _w > _h else _h

            ann_bbox_info = _generate_ann_bbox_info(_x, _y, _w, _h, area, ratio, _short, _long)
            ann_bbox_info.pop("x")
            ann_bbox_info.pop("y")

            for prop, val in ann_bbox_info.items():
                prop_stats = bbox_label_stats[prop]
                self._compute_far_from_mean(prop_stats, val, item_key, ann)

            for attr, value in ann.attributes.items():
                if attr in valid_attrs:
                    bbox_attr_stats = dist_by_attr[label_name][attr]
                    bbox_val_stats = bbox_attr_stats[str(value)]

                    for prop, val in ann_bbox_info.items():
                        prop_stats = bbox_val_stats[prop]
                        self._compute_far_from_mean(prop_stats, val, item_key, ann)

        # Compute far_from_mean from property
        for item_key, annotations in self.items:
            for ann in annotations:
                if _is_valid_bbox(item_key, ann):
                    _update_bbox_props_far_from_mean(item_key, ann)

        return stats

    def generate_reports(self, stats):
        """
        Validates the dataset for detection tasks based on its statistics.

        Parameters
        ----------
        dataset : IDataset object
        stats : Dict object

        Returns
        -------
        reports (list): List of validation reports (DatasetValidationError).
        """

        reports = self._generate_common_reports(stats)
        reports += self._check_negative_length(stats)
        reports += self._check_invalid_value(stats)

        defined_attr_dist = stats["attribute_distribution"]["defined_attributes"]

        dist_by_label = stats["point_distribution_in_label"]
        dist_by_attr = stats["point_distribution_in_attribute"]

        defined_labels = defined_attr_dist.keys()
        for label_name in defined_labels:
            bbox_label_stats = dist_by_label[label_name]
            bbox_attr_label = dist_by_attr.get(label_name, {})

            reports += self._check_far_from_label_mean(label_name, bbox_label_stats)
            reports += self._check_imbalanced_dist_in_label(label_name, bbox_label_stats)

            for attr_name, bbox_attr_stats in bbox_attr_label.items():
                reports += self._check_far_from_attr_mean(label_name, attr_name, bbox_attr_stats)
                reports += self._check_imbalanced_dist_in_attr(
                    label_name, attr_name, bbox_attr_stats
                )

        return reports

    def _update_prop_distributions(self, curr_stats, target_stats):
        for prop, val in curr_stats.items():
            prop_stats = target_stats[prop]
            prop_stats["distribution"].append(val)

    def _compute_prop_dist(self, label_categories, stats, update_stats_by_label):
        dist_by_label = stats["point_distribution_in_label"]
        dist_by_attr = stats["point_distribution_in_attribute"]
        point_dist_in_item = stats["point_distribution_in_dataset_item"]

        base_valid_attrs = label_categories.attributes

        for item_key, annotations in self.items:
            ann_count = len(annotations)
            point_dist_in_item[item_key] = ann_count

            for ann in annotations:
                if not 0 <= ann.label < len(label_categories):
                    label_name = ann.label
                    valid_attrs = set()
                else:
                    label_name = label_categories[ann.label].name
                    valid_attrs = base_valid_attrs.union(label_categories[ann.label].attributes)

                    point_label_stats = dist_by_label.setdefault(
                        label_name, deepcopy(self.point_template)
                    )
                    ann_point_info, _has_error = update_stats_by_label(
                        item_key, ann, point_label_stats
                    )

                for attr, value in ann.attributes.items():
                    if attr in valid_attrs:
                        point_attr_label = dist_by_attr.setdefault(label_name, {})
                        point_attr_stats = point_attr_label.setdefault(attr, {})
                        point_val_stats = point_attr_stats.setdefault(
                            str(value), deepcopy(self.point_template)
                        )

                        if not _has_error:
                            self._update_prop_distributions(ann_point_info, point_val_stats)

    def _compute_prop_stats_from_dist(self, dist_by_label, dist_by_attr):
        for label_name, stats in dist_by_label.items():
            prop_stats_list = list(stats.values())
            attr_label = dist_by_attr.get(label_name, {})
            for vals in attr_label.values():
                for val_stats in vals.values():
                    prop_stats_list += list(val_stats.values())

            for prop_stats in prop_stats_list:
                prop_dist = prop_stats.pop("distribution", [])
                if len(prop_dist) > 0:
                    prop_stats["mean"] = np.mean(prop_dist)
                    prop_stats["stdev"] = np.std(prop_dist)
                    prop_stats["min"] = np.min(prop_dist)
                    prop_stats["max"] = np.max(prop_dist)
                    prop_stats["median"] = np.median(prop_dist)

                    counts, bins = np.histogram(prop_dist)
                    prop_stats["histogram"]["bins"] = bins.tolist()
                    prop_stats["histogram"]["counts"] = counts.tolist()

    def _compute_far_from_mean(self, prop_stats, val, item_key, ann):
        def _far_from_mean(val, mean, stdev):
            thr = self.far_from_mean_thr
            return val > mean + (thr * stdev) or val < mean - (thr * stdev)

        mean = prop_stats["mean"]
        stdev = prop_stats["stdev"]

        if _far_from_mean(val, mean, stdev):
            items_far_from_mean = prop_stats["items_far_from_mean"]
            far_from_mean = items_far_from_mean.setdefault(item_key, {})
            far_from_mean[ann.id] = val

    def _check_negative_length(self, stats):
        validation_reports = []

        items_w_neg_len = stats["items_with_negative_length"]
        for item_dets, anns_w_neg_len in items_w_neg_len.items():
            item_id, item_subset = item_dets
            for ann_id, props in anns_w_neg_len.items():
                for prop, val in props.items():
                    val = round(val, 2)
                    details = (item_subset, ann_id, f"{self.str_ann_type} {prop}", val)
                    validation_reports += self._generate_validation_report(
                        NegativeLength, Severity.error, item_id, *details
                    )

        return validation_reports

    def _check_invalid_value(self, stats):
        validation_reports = []

        items_w_invalid_val = stats["items_with_invalid_value"]
        for item_dets, anns_w_invalid_val in items_w_invalid_val.items():
            item_id, item_subset = item_dets
            for ann_id, props in anns_w_invalid_val.items():
                for prop in props:
                    details = (item_subset, ann_id, f"{self.str_ann_type} {prop}")
                    validation_reports += self._generate_validation_report(
                        InvalidValue, Severity.error, item_id, *details
                    )

        return validation_reports

    def _check_imbalanced_dist_in_label(self, label_name, label_stats):
        validation_reports = []
        thr = self.dominance_thr
        topk_ratio = self.topk_bins_ratio

        for prop, prop_stats in label_stats.items():
            value_counts = prop_stats["histogram"]["counts"]
            n_bucket = len(value_counts)
            if n_bucket < 2:
                continue
            topk = max(1, int(np.around(n_bucket * topk_ratio)))

            if topk > 0:
                topk_values = np.sort(value_counts)[-topk:]
                ratio = np.sum(topk_values) / np.sum(value_counts)
                if ratio >= thr:
                    details = (label_name, f"{self.str_ann_type} {prop}")
                    validation_reports += self._generate_validation_report(
                        ImbalancedDistInLabel, Severity.info, *details
                    )

        return validation_reports

    def _check_imbalanced_dist_in_attr(self, label_name, attr_name, attr_stats):
        validation_reports = []
        thr = self.dominance_thr
        topk_ratio = self.topk_bins_ratio

        for attr_value, value_stats in attr_stats.items():
            for prop, prop_stats in value_stats.items():
                value_counts = prop_stats["histogram"]["counts"]
                n_bucket = len(value_counts)
                if n_bucket < 2:
                    continue
                topk = max(1, int(np.around(n_bucket * topk_ratio)))

                if topk > 0:
                    topk_values = np.sort(value_counts)[-topk:]
                    ratio = np.sum(topk_values) / np.sum(value_counts)
                    if ratio >= thr:
                        details = (label_name, attr_name, attr_value, f"{self.str_ann_type} {prop}")
                        validation_reports += self._generate_validation_report(
                            ImbalancedDistInAttribute, Severity.info, *details
                        )

        return validation_reports

    def _check_far_from_label_mean(self, label_name, label_stats):
        validation_reports = []

        for prop, prop_stats in label_stats.items():
            items_far_from_mean = prop_stats["items_far_from_mean"]
            if prop_stats["mean"] is not None:
                mean = round(prop_stats["mean"], 2)

            for item_dets, anns_far in items_far_from_mean.items():
                item_id, item_subset = item_dets
                for ann_id, val in anns_far.items():
                    val = round(val, 2)
                    details = (
                        item_subset,
                        label_name,
                        ann_id,
                        f"{self.str_ann_type} {prop}",
                        mean,
                        val,
                    )
                    validation_reports += self._generate_validation_report(
                        FarFromLabelMean, Severity.warning, item_id, *details
                    )

        return validation_reports

    def _check_far_from_attr_mean(self, label_name, attr_name, attr_stats):
        validation_reports = []

        for attr_value, value_stats in attr_stats.items():
            for prop, prop_stats in value_stats.items():
                items_far_from_mean = prop_stats["items_far_from_mean"]
                if prop_stats["mean"] is not None:
                    mean = round(prop_stats["mean"], 2)

                for item_dets, anns_far in items_far_from_mean.items():
                    item_id, item_subset = item_dets
                    for ann_id, val in anns_far.items():
                        val = round(val, 2)
                        details = (
                            item_subset,
                            label_name,
                            ann_id,
                            attr_name,
                            attr_value,
                            f"{self.str_ann_type} {prop}",
                            mean,
                            val,
                        )
                        validation_reports += self._generate_validation_report(
                            FarFromAttrMean, Severity.warning, item_id, *details
                        )

        return validation_reports


class SegmentationValidator(DetectionValidator):
    """
    A specific validator class for (instance) segmentation task.
    """

    def __init__(
        self,
        task_type=TaskType.segmentation,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        super().__init__(
            task_type=task_type,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_ratio_thr=dominance_ratio_thr,
            topk_bins=topk_bins,
        )

        self.point_template = {
            "area": deepcopy(self.numerical_stat_template),
            "width": deepcopy(self.numerical_stat_template),
            "height": deepcopy(self.numerical_stat_template),
        }

    def compute_statistics(self, dataset):
        """
        Computes statistics of the dataset for the segmentation task.

        Parameters
        ----------
        dataset : IDataset object

        Returns
        -------
        stats (dict): A dict object containing statistics of the dataset.
        """

        stats, filtered_items = self._compute_common_statistics(dataset)

        stats["items_with_invalid_value"] = {}
        stats["point_distribution_in_label"] = {}
        stats["point_distribution_in_attribute"] = {}
        stats["point_distribution_in_dataset_item"] = {}

        self.items = filtered_items

        def _generate_ann_mask_info(area, _w, _h):
            return {
                "area": area,
                "width": _w,
                "height": _h,
            }

        def _update_mask_stats_by_label(item_key, ann, mask_label_stats):
            mask_has_error = False

            _, _, _w, _h = ann.get_bbox()

            # Detete the following block when #226 is resolved
            # https://github.com/openvinotoolkit/datumaro/issues/226
            if ann.type == AnnotationType.mask:
                _w += 1
                _h += 1

            area = ann.get_area()

            ann_mask_info = _generate_ann_mask_info(area, _w, _h)

            items_w_invalid_val = stats["items_with_invalid_value"]
            for prop, val in ann_mask_info.items():
                if val == float("inf") or np.isnan(val):
                    mask_has_error = True
                    anns_w_invalid_val = items_w_invalid_val.setdefault(item_key, {})
                    invalid_props = anns_w_invalid_val.setdefault(ann.id, [])
                    invalid_props.append(prop)

            if not mask_has_error:
                self._update_prop_distributions(ann_mask_info, mask_label_stats)

            return ann_mask_info, mask_has_error

        # Collect property distribution
        label_categories = dataset.categories().get(AnnotationType.label, LabelCategories())
        self._compute_prop_dist(label_categories, stats, _update_mask_stats_by_label)

        # Compute property statistics from distribution
        dist_by_label = stats["point_distribution_in_label"]
        dist_by_attr = stats["point_distribution_in_attribute"]
        self._compute_prop_stats_from_dist(dist_by_label, dist_by_attr)

        def _is_valid_mask(item_key, ann):
            has_defined_label = 0 <= ann.label < len(label_categories)
            if not has_defined_label:
                return False

            mask_has_invalid_val = ann.id in stats["items_with_invalid_value"].get(item_key, {})
            return not mask_has_invalid_val

        def _update_mask_props_far_from_mean(item_key, ann):
            base_valid_attrs = label_categories.attributes
            valid_attrs = base_valid_attrs.union(label_categories[ann.label].attributes)
            label_name = label_categories[ann.label].name
            mask_label_stats = dist_by_label[label_name]

            _, _, _w, _h = ann.get_bbox()

            # Detete the following block when #226 is resolved
            # https://github.com/openvinotoolkit/datumaro/issues/226
            if ann.type == AnnotationType.mask:
                _w += 1
                _h += 1
            area = ann.get_area()

            ann_mask_info = _generate_ann_mask_info(area, _w, _h)

            for prop, val in ann_mask_info.items():
                prop_stats = mask_label_stats[prop]
                self._compute_far_from_mean(prop_stats, val, item_key, ann)

            for attr, value in ann.attributes.items():
                if attr in valid_attrs:
                    mask_attr_stats = dist_by_attr[label_name][attr]
                    mask_val_stats = mask_attr_stats[str(value)]

                    for prop, val in ann_mask_info.items():
                        prop_stats = mask_val_stats[prop]
                        self._compute_far_from_mean(prop_stats, val, item_key, ann)

        for item_key, annotations in self.items:
            for ann in annotations:
                if _is_valid_mask(item_key, ann):
                    _update_mask_props_far_from_mean(item_key, ann)

        return stats

    def generate_reports(self, stats):
        """
        Validates the dataset for segmentation tasks based on its statistics.

        Parameters
        ----------
        dataset : IDataset object
        stats : Dict object

        Returns
        -------
        reports (list): List of validation reports (DatasetValidationError).
        """

        reports = self._generate_common_reports(stats)
        reports += self._check_invalid_value(stats)

        defined_attr_dist = stats["attribute_distribution"]["defined_attributes"]

        dist_by_label = stats["point_distribution_in_label"]
        dist_by_attr = stats["point_distribution_in_attribute"]

        defined_labels = defined_attr_dist.keys()
        for label_name in defined_labels:
            mask_label_stats = dist_by_label[label_name]
            mask_attr_label = dist_by_attr.get(label_name, {})

            reports += self._check_far_from_label_mean(label_name, mask_label_stats)
            reports += self._check_imbalanced_dist_in_label(label_name, mask_label_stats)

            for attr_name, mask_attr_stats in mask_attr_label.items():
                reports += self._check_far_from_attr_mean(label_name, attr_name, mask_attr_stats)
                reports += self._check_imbalanced_dist_in_attr(
                    label_name, attr_name, mask_attr_stats
                )

        return reports


from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, List, Optional

from nltk.corpus import stopwords


@dataclass
class TabularValidationStats:
    total_ann_count: int = field(default=0)
    items_missing_annotation: List[Any] = field(default_factory=list)

    @classmethod
    def create_with_dataset(cls, dataset):
        instance = cls()
        instance.__post_init__(dataset)
        return instance

    def __post_init__(self, dataset: Optional[Any] = None):
        if dataset:
            self.label_categories = dataset.categories().get(
                AnnotationType.label, LabelCategories()
            )
            self.tabular_categories = dataset.categories().get(
                AnnotationType.caption, TabularCategories()
            )
            self.label_columns = list({item.parent for item in self.label_categories.items})
            self.caption_columns = [cat.name for cat in self.tabular_categories]

            self.defined_labels = {cat.name: 0 for cat in self.label_categories}
            self.empty_labels = {cat: [] for cat in self.label_columns}

            self.defined_captions = {cat: 0 for cat in self.caption_columns}
            self.empty_captions = {cat: [] for cat in self.caption_columns}
            self.redundancies = {
                cat: {"stopword": [], "url": [], "html": [], "emoji": []}
                for cat in self.caption_columns
            }

    def to_dict(self):
        empty_labels = self._build_empty_labels_dict(self.empty_labels, "items_with_empty_label")
        empty_captions = self._build_empty_labels_dict(
            self.empty_captions, "items_with_empty_caption"
        )
        redundancies = self._build_redundancies_dict(self.redundancies)

        return {
            "total_ann_count": self.total_ann_count,
            "items_missing_annotation": self.items_missing_annotation,
            "items_broken_annotation": self._collect_broken_annotations(),
            "label_distribution": {
                "defined_labels": self.defined_labels,
                "empty_labels": empty_labels,
            },
            "caption_distribution": {
                "defined_captions": self.defined_captions,
                "empty_captions": empty_captions,
                "redundancies": redundancies,
            },
        }

    def _build_empty_labels_dict(self, empty_dict, key_name):
        result = defaultdict(dict)
        for label, items in empty_dict.items():
            result[label]["count"] = len(items)
            result[label][key_name] = list(items)
        return result

    def _build_redundancies_dict(self, redundancies):
        result = defaultdict(lambda: defaultdict(dict))
        for caption, items in redundancies.items():
            for key in ["stopword", "url", "html", "emoji"]:
                result[caption][key]["count"] = len(items[key])
                result[caption][key]["items_with_redundancies"] = list(items[key])
        return result

    def _collect_broken_annotations(self):
        broken_annotations = set()
        for items in [self.empty_labels, self.empty_captions]:
            for key, value in items.items():
                broken_annotations.update(value)
        return list(broken_annotations)


class TabularValidator(_TaskValidator):
    """
    A specific validator class for tabular dataset.
    """

    def __init__(
        self,
        task_type=TaskType.tabular,
        few_samples_thr=None,
        imbalance_ratio_thr=None,
        far_from_mean_thr=None,
        dominance_ratio_thr=None,
        topk_bins=None,
    ):
        super().__init__(
            task_type=task_type,
            few_samples_thr=few_samples_thr,
            imbalance_ratio_thr=imbalance_ratio_thr,
            far_from_mean_thr=far_from_mean_thr,
            dominance_ratio_thr=dominance_ratio_thr,
            topk_bins=topk_bins,
        )

        self.numerical_stat_template = {
            "items_far_from_mean": {},
            "mean": None,
            "stdev": None,
            "min": None,
            "max": None,
            "median": None,
            "histogram": {
                "bins": [],
                "counts": [],
            },
            "distribution": [],
            "items_outlier": {},
            "outlier": None,
        }
        self.value_template = {"value": deepcopy(self.numerical_stat_template)}

    def _compute_common_statistics(self, dataset):
        stats = TabularValidationStats.create_with_dataset(dataset=dataset)

        filtered_items = []
        for item in dataset:
            item_key = (item.id, item.subset)
            label_check = {cat: 0 for cat in stats.label_columns}
            annotations = [ann for ann in item.annotations if ann.type in self.ann_types]
            ann_count = len(annotations)
            filtered_items.append((item_key, annotations))
            if ann_count == 0:
                stats.items_missing_annotation.append(item_key)
            stats.total_ann_count += ann_count

            caption_check = deepcopy(stats.caption_columns)
            for ann in annotations:
                if ann.type == AnnotationType.caption:
                    caption_ = ann.caption
                    for cat in stats.caption_columns:
                        if caption_.startswith(cat):
                            stats.defined_captions[cat] += 1
                            caption_ = caption_.split(cat + ":")[-1]
                            caption_check.remove(cat)
                            self._check_contain_redundancies(caption_, stats, cat, item_key)
                else:
                    label_name = stats.label_categories[ann.label].name
                    stats.defined_labels[label_name] += 1
                    label_name = label_name.split(":")[0]
                    label_check[label_name] += 1

            for cap in caption_check:
                stats.empty_captions[cap].append(item_key)

            for label_col, v in label_check.items():
                if v == 0:
                    stats.empty_labels[label_col].append(item_key)

        return stats.to_dict(), filtered_items

    def _check_contain_redundancies(self, text, stats, column, item_key):
        if column not in stats.redundancies.keys():
            return

        import re

        try:
            stop_words = set(stopwords.words("english"))  # TODO
        except LookupError:
            import nltk

            nltk.download("stopwords")
            stop_words = set(stopwords.words("english"))  # TODO

        def contains_emoji(text):
            return bool(emoji_pattern.search(text))

        def contains_html_tags(text):
            html_pattern = re.compile(r"<.*?>")
            return bool(html_pattern.search(text))

        def contains_url(text):
            url_pattern = re.compile(r"http\S+|www\S+|https\S+")
            return bool(url_pattern.search(text))

        def contains_stopword(text):
            return any(c in stop_words for c in str(text).lower())

        redun_stats = stats.redundancies[column]
        if contains_emoji(text):
            redun_stats["emoji"].append(item_key)
        if contains_html_tags(text):
            redun_stats["html"].append(item_key)
        if contains_url(text):
            redun_stats["url"].append(item_key)
        if contains_stopword(text):
            redun_stats["stopword"].append(item_key)

    def _compute_prop_dist(self, caption_columns, stats):
        dist_by_caption = stats["distribution_in_caption"]
        dist_in_item = stats["distribution_in_dataset_item"]

        for item_key, annotations in self.items:
            ann_count = len(annotations)
            dist_in_item[item_key] = ann_count

            for ann in annotations:
                if ann.type == AnnotationType.caption:
                    caption_ = ann.caption
                    for cat_name, type_ in caption_columns:
                        if caption_.startswith(cat_name):
                            caption_value = type_(caption_.split(f"{cat_name}:")[-1])
                            dist_by_caption[cat_name]["value"]["distribution"].append(caption_value)

    def _compute_prop_stats_from_dist(self, dist_by_caption):
        for stats in dist_by_caption.values():
            prop_stats = list(stats.values())[0]

            prop_dist = prop_stats.pop("distribution", [])
            if prop_dist:
                prop_stats["mean"] = np.mean(prop_dist)
                prop_stats["stdev"] = np.std(prop_dist)
                prop_stats["min"] = np.min(prop_dist)
                prop_stats["max"] = np.max(prop_dist)
                prop_stats["median"] = np.median(prop_dist)

                counts, bins = np.histogram(prop_dist)
                prop_stats["histogram"]["bins"] = bins.tolist()
                prop_stats["histogram"]["counts"] = counts.tolist()

                # Calculate Q1 (25th percentile) and Q3 (75th percentile)
                Q1, Q3 = np.percentile(prop_dist, [25, 75])
                IQR = Q3 - Q1

                # Calculate the acceptable range
                lower_bound = Q1 - 1.5 * IQR
                upper_bound = Q3 + 1.5 * IQR
                prop_stats["outlier"] = (lower_bound, upper_bound)

    def _compute_far_from_mean(self, prop_stats, val, item_key):
        def _far_from_mean(val, mean, stdev):
            thr = self.far_from_mean_thr
            return val > mean + (thr * stdev) or val < mean - (thr * stdev)

        mean = prop_stats["mean"]
        stdev = prop_stats["stdev"]

        if _far_from_mean(val, mean, stdev):
            prop_stats["items_far_from_mean"][item_key] = val

    def _compute_outlier(self, prop_stats, val, item_key):
        lower_bound, upper_bound = prop_stats["outlier"]

        if (val < lower_bound) | (val > upper_bound):
            prop_stats["items_outlier"][item_key] = val

    def compute_statistics(self, dataset):
        """
        Computes statistics of the tabular dataset.

        Parameters
        ----------
        dataset : IDataset object

        Returns
        -------
        stats (dict): A dict object containing statistics of the dataset.
        """

        stats, filtered_items = self._compute_common_statistics(dataset)

        self.items = filtered_items

        num_caption_columns = [
            (cat.name, cat.dtype)
            for cat in dataset.categories().get(AnnotationType.caption, TabularCategories())
            if cat.dtype in [int, float]
        ]

        stats["distribution_in_caption"] = {
            cap: deepcopy(self.value_template) for cap, _ in num_caption_columns
        }
        stats["distribution_in_dataset_item"] = {}

        # Collect property distribution
        self._compute_prop_dist(num_caption_columns, stats)

        # Compute property statistics from distribution
        dist_by_caption = stats["distribution_in_caption"]
        self._compute_prop_stats_from_dist(dist_by_caption)

        def _update_captions_far_from_mean_outlier(caption_columns, item_key, ann):
            for col, type_ in caption_columns:
                prop_stats = dist_by_caption[col]["value"]
                if ann.caption.startswith(col):
                    val = type_(ann.caption.split(f"{col}:")[-1])
                    self._compute_far_from_mean(prop_stats, val, item_key)
                    self._compute_outlier(prop_stats, val, item_key)

        # Compute far_from_mean and outlier from property
        for item_key, annotations in self.items:
            for ann in annotations:
                if ann.type == AnnotationType.caption:
                    _update_captions_far_from_mean_outlier(num_caption_columns, item_key, ann)

        return stats

    def generate_reports(self, stats):
        """
        Validates the dataset for classification tasks based on its statistics.

        Parameters
        ----------
        dataset : IDataset object
        stats: Dict object

        Returns
        -------
        reports (list): List of validation reports (DatasetValidationError).
        """
        reports = []

        # report for dataset
        reports += self._check_missing_label_categories(stats)

        # report for item
        reports += self._check_missing_annotation(stats)

        # report for label
        reports += self._check_few_samples_in_label(stats)
        reports += self._check_imbalanced_labels(stats)

        # report for caption
        reports += self._check_few_samples_in_caption(stats)
        reports += self._check_redundancies_in_caption(stats)
        reports += self._check_imbalanced_captions(stats)

        # report for missing value
        reports += self._check_broken_annotation(stats)
        reports += self._check_empty_label(stats)
        reports += self._check_empty_caption(stats)

        dist_by_caption = stats["distribution_in_caption"]
        for caption, caption_stats in dist_by_caption.items():
            reports += self._check_far_from_caption_mean(caption, caption_stats)
            reports += self._check_caption_outliers(caption, caption_stats)
            reports += self._check_imbalanced_dist_in_caption(caption, caption_stats)

        return reports

    def _check_broken_annotation(self, stats):
        validation_reports = []

        items_broken = stats["items_broken_annotation"]
        for item_id, item_subset in items_broken:
            validation_reports += self._generate_validation_report(
                BrokenAnnotation, Severity.warning, item_id, item_subset, self.str_ann_type
            )

        return validation_reports

    def _check_empty_label(self, stats):
        validation_reports = []

        empty_label_dist = stats["label_distribution"]["empty_labels"]
        for label_name, label_stats in empty_label_dist.items():
            for item_id, item_subset in label_stats["items_with_empty_label"]:
                details = (item_subset, label_name)
                validation_reports += self._generate_validation_report(
                    EmptyLabel, Severity.warning, item_id, *details
                )

        return validation_reports

    def _check_empty_caption(self, stats):
        validation_reports = []

        empty_caption_dist = stats["caption_distribution"]["empty_captions"]
        for caption_name, caption_stats in empty_caption_dist.items():
            for item_id, item_subset in caption_stats["items_with_empty_caption"]:
                details = (item_subset, caption_name)
                validation_reports += self._generate_validation_report(
                    EmptyCaption, Severity.warning, item_id, *details
                )

        return validation_reports

    def _check_few_samples_in_caption(self, stats):
        validation_reports = []
        thr = self.few_samples_thr

        defined_caption_dist = stats["caption_distribution"]["defined_captions"]
        captions_with_few_samples = [
            (caption_name, count)
            for caption_name, count in defined_caption_dist.items()
            if 0 < count <= thr
        ]

        for caption_name, count in captions_with_few_samples:
            validation_reports += self._generate_validation_report(
                FewSamplesInCaption, Severity.info, caption_name, count
            )

        return validation_reports

    def _check_far_from_caption_mean(self, caption_name, caption_stats):
        prop_stats = list(caption_stats.values())[0]

        if prop_stats["mean"] is not None:
            mean = round(prop_stats["mean"], 2)
        stdev = prop_stats["stdev"]
        upper_bound = mean + (self.far_from_mean_thr * stdev)
        lower_bound = mean - (self.far_from_mean_thr * stdev)

        validation_reports = []
        items_far_from_mean = prop_stats["items_far_from_mean"]
        for item_dets, val in items_far_from_mean.items():
            item_id, item_subset = item_dets
            val = round(val, 2)
            details = (
                item_subset,
                caption_name,
                mean,
                upper_bound,
                lower_bound,
                val,
            )
            validation_reports += self._generate_validation_report(
                FarFromCaptionMean, Severity.info, item_id, *details
            )

        return validation_reports

    def _check_caption_outliers(self, caption_name, caption_stats):
        prop_stats = list(caption_stats.values())[0]
        lower_bound, upper_bound = prop_stats["outlier"]
        items_outlier = prop_stats["items_outlier"]

        validation_reports = []
        for item_dets, val in items_outlier.items():
            item_id, item_subset = item_dets
            val = round(val, 2)
            details = (
                item_subset,
                caption_name,
                lower_bound,
                upper_bound,
                val,
            )
            validation_reports += self._generate_validation_report(
                OutlierInCaption, Severity.info, item_id, *details
            )

        return validation_reports

    def _check_redundancies_in_caption(self, stats):
        validation_reports = []

        redundancies_in_caption_dist = stats["caption_distribution"]["redundancies"]
        captions_with_redundancies = []
        for cap_column, cap_stats in redundancies_in_caption_dist.items():
            for redundancy_type, items in cap_stats.items():
                if 0 < items["count"]:
                    captions_with_redundancies.append((cap_column, redundancy_type, items["count"]))

        for cap_column, redundancy_type, count in captions_with_redundancies:
            validation_reports += self._generate_validation_report(
                RedundanciesInCaption, Severity.info, cap_column, redundancy_type, count
            )

        return validation_reports

    def _check_imbalanced_captions(self, stats):
        validation_reports = []
        thr = self.imbalance_ratio_thr

        defined_caption_dist = stats["caption_distribution"]["defined_captions"]
        count_by_caption_labels = [count for _, count in defined_caption_dist.items()]

        if len(defined_caption_dist) == 0:
            return validation_reports

        count_max = np.max(count_by_caption_labels)
        count_min = np.min(count_by_caption_labels)
        balance = count_max / count_min if count_min > 0 else float("inf")
        if balance >= thr:
            validation_reports += self._generate_validation_report(
                ImbalancedCaptions, Severity.info
            )

        return validation_reports

    def _check_imbalanced_dist_in_caption(self, caption_name, caption_stats):
        validation_reports = []
        thr = self.dominance_thr
        topk_ratio = self.topk_bins_ratio

        for prop, prop_stats in caption_stats.items():
            value_counts = prop_stats["histogram"]["counts"]
            n_bucket = len(value_counts)
            if n_bucket < 2:
                continue
            topk = max(1, int(np.around(n_bucket * topk_ratio)))

            if topk > 0:
                topk_values = np.sort(value_counts)[-topk:]
                ratio = np.sum(topk_values) / np.sum(value_counts)
                if ratio >= thr:
                    validation_reports += self._generate_validation_report(
                        ImbalancedDistInCaption, Severity.info, caption_name
                    )

        return validation_reports



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT

import codecs
import hashlib
import logging as log
import os
import os.path as osp
import string
from collections import OrderedDict

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.errors import DatasetExportError
from datumaro.components.exporter import Exporter
from datumaro.components.lazy_plugin import extra_deps
from datumaro.components.media import Image, ImageFromBytes
from datumaro.util.annotation_util import find_group_leader, find_instances, max_bbox
from datumaro.util.image import encode_image
from datumaro.util.mask_tools import merge_masks
from datumaro.util.tf_util import import_tf as _import_tf

from .format import DetectionApiPath

tf = _import_tf()


# filter out non-ASCII characters, otherwise training will crash
_printable = set(string.printable)


def _make_printable(s):
    return "".join(filter(lambda x: x in _printable, s))


def int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def int64_list_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))


def bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def bytes_list_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def float_list_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


@extra_deps("tensorflow")
class TfDetectionApiExporter(Exporter):
    DEFAULT_IMAGE_EXT = DetectionApiPath.DEFAULT_IMAGE_EXT

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--save-masks",
            action="store_true",
            help="Include instance masks (default: %(default)s)",
        )
        return parser

    def __init__(self, extractor, save_dir, save_masks=False, **kwargs):
        super().__init__(extractor, save_dir, **kwargs)

        self._save_masks = save_masks

    def _apply_impl(self):
        os.makedirs(self._save_dir, exist_ok=True)

        label_categories = self._extractor.categories().get(AnnotationType.label, LabelCategories())
        get_label = (
            lambda label_id: label_categories.items[label_id].name if label_id is not None else ""
        )
        label_ids = OrderedDict(
            (label.name, 1 + idx) for idx, label in enumerate(label_categories.items)
        )
        map_label_id = lambda label_id: label_ids.get(get_label(label_id), 0)
        self._get_label = get_label
        self._get_label_id = map_label_id

        for subset_name, subset in self._extractor.subsets().items():
            labelmap_path = osp.join(self._save_dir, DetectionApiPath.LABELMAP_FILE)
            with codecs.open(labelmap_path, "w", encoding="utf8") as f:
                for label, idx in label_ids.items():
                    f.write(
                        "item {\n" + ("\tid: %s\n" % (idx)) + ("\tname: '%s'\n" % (label)) + "}\n\n"
                    )

            anno_path = osp.join(self._save_dir, "%s.tfrecord" % (subset_name))
            with tf.io.TFRecordWriter(anno_path) as writer:
                for item in subset:
                    tf_example = self._make_tf_example(item)
                    writer.write(tf_example.SerializeToString())

    @staticmethod
    def _find_instances(annotations):
        return find_instances(
            a for a in annotations if a.type in {AnnotationType.bbox, AnnotationType.mask}
        )

    def _find_instance_parts(self, group, img_width, img_height):
        boxes = [a for a in group if a.type == AnnotationType.bbox]
        masks = [a for a in group if a.type == AnnotationType.mask]

        anns = boxes + masks
        leader = find_group_leader(anns)
        bbox = max_bbox(anns)

        mask = None
        if self._save_masks:
            mask = merge_masks(m.image for m in masks)

        return [leader, mask, bbox]

    def _export_instances(self, instances, width, height):
        xmins = []  # List of normalized left x coordinates of bounding boxes (1 per box)
        xmaxs = []  # List of normalized right x coordinates of bounding boxes (1 per box)
        ymins = []  # List of normalized top y coordinates of bounding boxes (1 per box)
        ymaxs = []  # List of normalized bottom y coordinates of bounding boxes (1 per box)
        classes_text = []  # List of class names of bounding boxes (1 per box)
        classes = []  # List of class ids of bounding boxes (1 per box)
        masks = []  # List of PNG-encoded instance masks (1 per box)

        for leader, mask, box in instances:
            label = _make_printable(self._get_label(leader.label))
            classes_text.append(label.encode("utf-8"))
            classes.append(self._get_label_id(leader.label))

            xmins.append(box[0] / width)
            xmaxs.append((box[0] + box[2]) / width)
            ymins.append(box[1] / height)
            ymaxs.append((box[1] + box[3]) / height)

            if self._save_masks:
                if mask is not None:
                    mask = encode_image(mask, ".png")
                else:
                    mask = b""
                masks.append(mask)

        result = {}
        if classes:
            result = {
                "image/object/bbox/xmin": float_list_feature(xmins),
                "image/object/bbox/xmax": float_list_feature(xmaxs),
                "image/object/bbox/ymin": float_list_feature(ymins),
                "image/object/bbox/ymax": float_list_feature(ymaxs),
                "image/object/class/text": bytes_list_feature(classes_text),
                "image/object/class/label": int64_list_feature(classes),
            }
            if masks:
                result["image/object/mask"] = bytes_list_feature(masks)
        return result

    def _make_tf_example(self, item):
        features = {
            "image/source_id": bytes_feature(
                str(item.attributes.get("source_id") or "").encode("utf-8")
            ),
        }

        filename = self._make_image_filename(item)
        features["image/filename"] = bytes_feature(filename.encode("utf-8"))

        if not isinstance(item.media, Image):
            raise DatasetExportError(
                "Failed to export dataset item '%s': " "item has no image info" % item.id
            )
        height, width = item.media.size

        features.update(
            {
                "image/height": int64_feature(height),
                "image/width": int64_feature(width),
            }
        )

        features.update(
            {
                "image/encoded": bytes_feature(b""),
                "image/format": bytes_feature(b""),
                "image/key/sha256": bytes_feature(b""),
            }
        )
        if self._save_media:
            if isinstance(item.media, Image) and item.media.has_data:
                buffer, fmt = self._save_image(item, filename)
                key = hashlib.sha256(buffer).hexdigest()

                features.update(
                    {
                        "image/encoded": bytes_feature(buffer),
                        "image/format": bytes_feature(fmt.encode("utf-8")),
                        "image/key/sha256": bytes_feature(key.encode("utf8")),
                    }
                )
            else:
                log.warning("Item '%s' has no image" % item.id)

        instances = self._find_instances(item.annotations)
        instances = [self._find_instance_parts(i, width, height) for i in instances]
        features.update(self._export_instances(instances, width, height))

        tf_example = tf.train.Example(features=tf.train.Features(feature=features))

        return tf_example

    def _save_image(self, item, path=None):  # pylint: disable=arguments-differ
        src_ext = item.media.ext.lower() if item.media.ext else item.media.ext
        dst_ext = osp.splitext(osp.basename(path))[1].lower()
        fmt = DetectionApiPath.IMAGE_EXT_FORMAT.get(dst_ext, "")
        if not fmt:
            log.warning(
                "Item '%s': can't find format string for the '%s' "
                "image extension, the corresponding field will be empty." % (item.id, dst_ext)
            )

        buffer = None
        if src_ext == dst_ext:
            if isinstance(item.media, ImageFromBytes):
                buffer = item.media.bytes
        if buffer is None:
            buffer = encode_image(item.media.data, dst_ext)
        return buffer, fmt

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        for subset in patch.updated_subsets:
            cls.convert(dataset.get_subset(subset), save_dir=save_dir, **kwargs)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# This model detects the bounding boxes of human faces in an image.
# It is provided by OpenVINO™ Model Server Quickstart guide,
# https://github.com/openvinotoolkit/model_server/blob/main/docs/ovms_quickstart.md

from typing import List, Tuple

import cv2
import numpy as np

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, AnnotationType, Bbox, LabelCategories
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image


class FaceDetectionModelInterpreter(IModelInterpreter):
    height = 400
    width = 600
    conf_threshold = 0.5

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        assert img.ndim == 3
        h_img, w_img, _ = img.shape

        # This should be stored to reconstruct the bbox points
        img_size = h_img, w_img

        img = cv2.resize(img, (self.width, self.height))
        img = img.transpose(2, 0, 1).astype(np.dtype("<f"))
        return img, img_size

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        detection = pred.get("detection_out")
        h_img, w_img = info  # Info has the image size

        bboxes = []
        # detection has (1, 200, 7) dimensions (200 detections for an image)
        for idx, det in enumerate(detection[0, :, :]):
            conf = det[2]
            if conf <= self.conf_threshold:
                continue

            label = det[1]
            # These values are in [0, 1], so that we have to rescale them.
            x_min, y_min = w_img * det[3], h_img * det[4]
            x_max, y_max = w_img * det[5], h_img * det[6]
            bboxes.append(
                Bbox(
                    x=x_min,
                    y=y_min,
                    w=x_max - x_min,
                    h=y_max - y_min,
                    id=idx,
                    label=label,
                )
            )
        return bboxes

    def get_categories(self):
        label_categories = LabelCategories()
        label_categories.add(name="empty")
        label_categories.add(name="face")
        return {AnnotationType.label: label_categories}



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import contextlib
import fnmatch
import glob
import logging as log
import os.path as osp
from dataclasses import dataclass, field
from enum import Enum, IntEnum, auto
from io import BufferedReader
from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Union,
)

from typing_extensions import Protocol


class FormatDetectionConfidence(IntEnum):
    """
    Represents the level of confidence that a detector has in a dataset
    belonging to the detector's format.
    """

    NONE = 1
    """
    """
    EXTREME_LOW = 5
    """
    EXTREME_LOW: This is currently only assigned to ImageDir format. This is because
    ImageDir format can be detected in every image dataset format.
    """
    LOW = 10
    """
    LOW: The dataset seems to belong to the format, but the format is too loosely
    defined to be able to distinguish it from other formats.
    """
    MEDIUM = 20
    """
    MEDIUM: The dataset seems to belong to the format, and is likely not to belong
    to any other format.
    """
    # There's no HIGH confidence yet, because none of the detectors
    # deserve it. It's reserved for when the detector is sure that
    # the dataset belongs to the format; for example, because the format
    # has explicit identification via magic numbers/files.


@dataclass(order=True, frozen=True)
class DetectedFormat:
    confidence: FormatDetectionConfidence = field(compare=True)
    name: str = field(compare=False)

    def __eq__(self, __o: "DetectedFormat") -> bool:
        return self.name == __o.name

    def __str__(self) -> str:
        return self.name


# All confidence levels should be positive for a couple of reasons:
# * It makes it possible to use 0 or a negative number as a special
#   value that is guaranteed to be less than any real value.
# * It makes sure that every confidence level is a true value.
assert all(level > 0 for level in FormatDetectionConfidence)


class RejectionReason(Enum):
    unmet_requirements = auto()
    insufficient_confidence = auto()
    detection_unsupported = auto()


class _FormatRejected(Exception):
    @property
    def reason(self) -> RejectionReason:
        raise NotImplementedError


class FormatRequirementsUnmet(_FormatRejected):
    """
    Represents a situation where a dataset does not meet the requirements
    of a given dataset format.
    More specifically, if this exception is raised, then it is necessary
    (but may not be sufficient) for the dataset to meet at least
    one of these requirements to be detected as being in that format.

    Each element of `failed_alternatives` must be a human-readable
    statement describing a requirement that was not met.

    Must not be constructed or raised directly; use `FormatDetectionContext`
    methods.
    """

    def __init__(self, failed_alternatives: Sequence[str]) -> None:
        assert failed_alternatives
        self.failed_alternatives = tuple(failed_alternatives)

    def __str__(self) -> str:
        lines = []

        if len(self.failed_alternatives) > 1:
            lines.append("None of the following requirements were met:")
        else:
            lines.append("The following requirement was not met:")

        lines.extend("  " + req for req in self.failed_alternatives)

        return "\n".join(lines)

    @property
    def reason(self) -> RejectionReason:
        return RejectionReason.unmet_requirements


class FormatDetectionUnsupported(_FormatRejected):
    """
    Represents a situation where detection is attempted with a format that
    does not support it.

    Must not be constructed or raised directly; use
    `FormatDetectionContext.raise_unsupported` instead.
    """

    def __str__(self) -> str:
        return "Detection for this format is unsupported"

    @property
    def reason(self) -> RejectionReason:
        return RejectionReason.detection_unsupported


class FormatDetectionContext:
    """
    An instance of this class is given to a dataset format detector.
    See the `FormatDetector` documentation. The class should not
    be instantiated directly.

    A context encapsulates information about the dataset whose format
    is being detected. It also offers methods that place requirements
    on that dataset. Each such method raises a `FormatRequirementsUnmet`
    exception if the requirement is not met. If the requirement _is_
    met, the return value depends on the method.
    """

    class _OneOrMoreContext:
        failed_alternatives: List[str]
        had_successful_alternatives: bool

        def __init__(self) -> None:
            self.failed_alternatives = []
            self.had_successful_alternatives = False

    # This points to a `_OneOrMoreContext` when and only when the detector
    # is directly within a `require_any` block.
    _one_or_more_context: Optional[_OneOrMoreContext]

    def __init__(self, root_path: str) -> None:
        self._root_path = root_path
        self._one_or_more_context = None

    @property
    def root_path(self) -> str:
        """
        Returns the path to the root directory of the dataset.
        Detectors should avoid using this property in favor of specific
        requirement methods.
        """
        return self._root_path

    def _is_path_within_root(self, path: str) -> bool:
        """
        Checks that `path` is a relative path and does not attempt to leave
        the dataset root by using `..` segments.

        Requirement-placing methods that use this to verify their arguments
        should raise a FormatRequirementsUnmet rather than a "hard" error like
        AssertionError if False is returned. The reason is that the path passed
        by the detector might not have been hardcoded, and instead might have
        been acquired from another file in the dataset. In that case, an invalid
        pattern signifies a problem with the dataset, not with the detector.
        """
        if osp.isabs(path) or osp.splitdrive(path)[0]:
            return False

        path = osp.normpath(path)
        if path.startswith(".." + osp.sep):
            return False

        return True

    def _start_requirement(self, req_type: str) -> None:
        assert not self._one_or_more_context, (
            f"a requirement ({req_type}) can't be placed directly within " "a 'require_any' block"
        )

    def raise_unsupported(self) -> NoReturn:
        """
        Raises a `FormatDetectionUnsupported` exception to signal that the
        current format does not support detection.
        """

        raise FormatDetectionUnsupported

    def fail(self, requirement_desc: str) -> NoReturn:
        """
        Places a requirement that is never met. `requirement_desc` must contain
        a human-readable description of the requirement.
        """
        self._start_requirement("fail")

        raise FormatRequirementsUnmet((requirement_desc,))

    def require_file(
        self,
        pattern: str,
        *,
        exclude_fnames: Union[str, Collection[str]] = (),
    ) -> str:
        """
        Places the requirement that the dataset contains at least one file whose
        relative path matches the given pattern. The pattern must be a glob-like
        pattern; `**` can be used to indicate a sequence of zero or more
        subdirectories.
        If the pattern does not describe a relative path, or refers to files
        outside the dataset root, the requirement is considered unmet.
        If the requirement is met, the relative path to one of the files that
        match the pattern is returned. If there are multiple such files, it's
        unspecified which one of them is returned.

        `exclude_fnames` must be a collection of patterns or a single pattern.
        If at least one pattern is supplied, then the placed requirement is
        narrowed to only accept files with names that match none of these
        patterns.
        """

        self._start_requirement("require_file")

        return next(self._require_files_iter(pattern, exclude_fnames=exclude_fnames))

    def require_files(
        self,
        pattern: str,
        *,
        exclude_fnames: Union[str, Collection[str]] = (),
    ) -> List[str]:
        """
        Same as `require_file`, but returns all matching paths in alphabetical
        order.
        """

        self._start_requirement("require_files")

        return sorted(self._require_files_iter(pattern, exclude_fnames=exclude_fnames))

    def require_files_iter(
        self,
        pattern: str,
        *,
        exclude_fnames: Union[str, Collection[str]] = (),
    ) -> Iterator[str]:
        """
        Same as `require_files`, but returns a generator.
        """

        self._start_requirement("require_files_iter")

        return self._require_files_iter(pattern, exclude_fnames=exclude_fnames)

    def _require_files_iter(
        self,
        pattern: str,
        *,
        exclude_fnames: Union[str, Collection[str]] = (),
    ) -> Iterator[str]:
        if isinstance(exclude_fnames, str):
            exclude_fnames = (exclude_fnames,)

        requirement_desc = f'dataset must contain a file matching pattern "{pattern}"'

        if exclude_fnames:
            requirement_desc += (
                " (but not named " + ", ".join(f'"{e}"' for e in exclude_fnames) + ")"
            )

        if not self._is_path_within_root(pattern):
            self.fail(requirement_desc)

        pattern_abs = osp.join(glob.escape(self._root_path), pattern)
        none_found = True
        for path in glob.iglob(pattern_abs, recursive=True):
            if osp.isfile(path):
                # Ideally, we should provide a way to filter out whole paths,
                # not just file names. However, there is no easy way to match an
                # entire path with a pattern (fnmatch is unsuitable, because
                # it lets '*' match a slash, which can lead to spurious matches
                # and is not how glob works).
                if any(fnmatch.fnmatch(osp.basename(path), pat) for pat in exclude_fnames):
                    continue

                yield osp.relpath(path, self._root_path)
                none_found = False

        if none_found:
            self.fail(requirement_desc)

    @contextlib.contextmanager
    def probe_text_file(
        self, path: str, requirement_desc: str, is_binary_file: bool = False
    ) -> Iterator[Union[BufferedReader, TextIO]]:
        """
        Returns a context manager that can be used to place a requirement on
        the contents of the file referred to by `path`. To do so, you must
        enter and exit this context manager (typically, by using the `with`
        statement). On entering, the file is opened for reading in text mode and
        the resulting file object is returned. On exiting, the file object is
        closed.

        The requirement that is placed by doing this is considered met if all
        of the following are true:

        * `path` is a relative path that refers to a file within the dataset
          root.
        * The file is opened successfully.
        * The context is exited without an exception.

        If the context is exited with an exception that was produced by another
        requirement being unmet, that exception is reraised and the new
        requirement is abandoned.

        `requirement_desc` must be a human-readable statement describing the
        requirement.
        """

        self._start_requirement("probe_text_file")

        requirement_desc_full = f"{path}: {requirement_desc}"

        if not self._is_path_within_root(path):
            self.fail(requirement_desc_full)

        try:
            if is_binary_file:
                with open(osp.join(self._root_path, path), "rb") as f:
                    yield f
            else:
                with open(osp.join(self._root_path, path), encoding="utf-8") as f:
                    yield f
        except _FormatRejected:
            raise
        except Exception:
            self.fail(requirement_desc_full)

    @contextlib.contextmanager
    def require_any(self) -> Iterator[None]:
        """
        Returns a context manager that can be used to place a requirement that
        is considered met if at least one of several alternative sets of
        requirements is met.
        To do so, use a `with` statement, with the alternative sets of
        requirements represented as nested `with` statements using the context
        manager returned by `alternative`:

        .. code-block::

            with context.require_any():
                with context.alternative():
                    # place requirements from alternative set 1 here
                with context.alternative():
                    # place requirements from alternative set 2 here
                ...

        The contents of all `with context.alternative()` blocks will be
        executed, even if an alternative that is met is found early.

        Requirements must not be placed directly within a
        `with context.require_any()` block.
        """

        self._start_requirement("require_any")

        self._one_or_more_context = self._OneOrMoreContext()

        try:
            yield

            # If at least one `alternative` block succeeded,
            # then the `require_any` block succeeds.
            if self._one_or_more_context.had_successful_alternatives:
                return

            # If no alternatives succeeded, and none failed, then there were
            # no alternatives at all.
            assert self._one_or_more_context.failed_alternatives, (
                "a 'require_any' block must contain " "at least one 'alternative' block"
            )

            raise FormatRequirementsUnmet(self._one_or_more_context.failed_alternatives)
        finally:
            self._one_or_more_context = None

    @contextlib.contextmanager
    def alternative(self) -> Iterator[None]:
        """
        Returns a context manager that can be used in combination with
        `require_any` to define alternative requirements. See the
        documentation for `require_any` for more details.

        Must only be used directly within a `with context.requirements()` block.
        """

        assert self._one_or_more_context, (
            "An 'alternative' block must be directly within " "a 'require_any' block"
        )

        saved_one_or_more_context = self._one_or_more_context
        self._one_or_more_context = None

        try:
            yield
        except FormatRequirementsUnmet as e:
            saved_one_or_more_context.failed_alternatives.extend(e.failed_alternatives)
        else:
            saved_one_or_more_context.had_successful_alternatives = True
        finally:
            self._one_or_more_context = saved_one_or_more_context


FormatDetector = Callable[
    [FormatDetectionContext],
    Optional[FormatDetectionConfidence],
]
"""
Denotes a callback that implements detection for a specific dataset format.
The callback receives an instance of `FormatDetectionContext` and must call
methods on that instance to place requirements that the dataset must meet
in order for it to be considered as belonging to the format.

Must terminate in one of the following ways:

* by returning the level of confidence in the dataset belonging to the format
  (or `None`, which is equivalent to the `MEDIUM` level);
* by raising a `FormatRequirementsUnmet` exception via one of
  the `FormatDetectionContext` methods;
* by raising a `FormatDetectionUnsupported` exception via
  `FormatDetectionContext.raise_unsupported`.
"""


def apply_format_detector(
    dataset_root_path: str,
    detector: FormatDetector,
) -> FormatDetectionConfidence:
    """
    Checks whether the dataset located at `dataset_root_path` belongs to the
    format detected by `detector`. If it does, returns the confidence level
    of the detection. Otherwise, terminates with the exception that was raised
    by the detector.
    """
    context = FormatDetectionContext(dataset_root_path)

    if not osp.isdir(dataset_root_path):
        context.fail(f"root path {dataset_root_path} must refer to a directory")

    return detector(context) or FormatDetectionConfidence.MEDIUM


class RejectionCallback(Protocol):
    def __call__(
        self,
        format_name: str,
        reason: RejectionReason,
        human_message: str,
    ) -> Any:
        ...


def detect_dataset_format(
    formats: Iterable[Tuple[str, FormatDetector]],
    path: str,
    *,
    rejection_callback: Optional[RejectionCallback] = None,
) -> Sequence[DetectedFormat]:
    """
    Determines which format(s) the dataset at the specified path belongs to.

    The function applies each supplied detector to the given patch and decides
    whether the corresponding format is detected or rejected. A format may be
    rejected if the detector fails or if it succeeds with less confidence than
    another detector (other rejection reasons might be added in the future).

    Args:
        `formats`: The formats to be considered. Each element of the
            iterable must be a tuple of a format name and a `FormatDetector`
            instance.

        `path`: the filesystem path to the dataset to be analyzed.

        `rejection_callback`: Unless `None`, called for every rejected format
            to report the reason it was rejected.

    Returns: a sequence of detected format names.
    """

    if not osp.exists(path):
        raise FileNotFoundError(f"Path {path} doesn't exist")

    def report_insufficient_confidence(
        format_name: str,
        format_with_more_confidence: str,
    ):
        if rejection_callback:
            rejection_callback(
                format_name,
                RejectionReason.insufficient_confidence,
                f"Another format ({format_with_more_confidence}) "
                "was matched with more confidence",
            )

    max_confidence = 0
    matches: List[DetectedFormat] = []

    for format_name, detector in formats:
        log.debug("Checking '%s' format...", format_name)
        try:
            new_confidence = apply_format_detector(path, detector)
        except _FormatRejected as ex:
            human_message = str(ex)
            if rejection_callback:
                rejection_callback(format_name, ex.reason, human_message)
            log.debug(human_message)
        else:
            log.debug("Format matched with confidence %d", new_confidence)

            # keep only matches with the highest confidence
            if new_confidence > max_confidence:
                for match in matches:
                    report_insufficient_confidence(match.name, format_name)

                matches = [DetectedFormat(new_confidence, format_name)]
                max_confidence = new_confidence
            elif new_confidence == max_confidence:
                matches.append(DetectedFormat(new_confidence, format_name))
            else:  # new confidence is less than max
                report_insufficient_confidence(format_name, matches[0].name)

    # TODO: This should be controlled by our priority logic.
    # However, some datasets' detect() are currently broken,
    # so that it is inevitable to introduce this.
    # We must revisit this after fixing detect().
    def _give_more_priority_to_with_subset_dirs(matches: List[DetectedFormat]):
        for idx, match in enumerate(matches):
            with_subset_dir_match = DetectedFormat(
                match.confidence, match.name + "_with_subset_dirs"
            )
            if with_subset_dir_match in matches:
                matches = matches.pop(idx)
                return True
        return False

    while _give_more_priority_to_with_subset_dirs(matches):
        continue

    return matches



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import math
import re
from typing import TYPE_CHECKING

from .algorithm import InferenceResultAnalyzer

if TYPE_CHECKING:
    import pandas as pd
else:
    from datumaro.util.import_util import lazy_import

    pd = lazy_import("pandas")


class SampleEntropy(InferenceResultAnalyzer):
    """
    Entropy is a class that inherits an Sampler,
    calculates an uncertainty score based on an entropy,
    and get samples based on that score.
    """

    def __init__(self, data, inference):
        """
        Constructor function

        Args:

            data: Receive the data format in pd.DataFrame format.
                    ImageID is an essential element for data.
            inference:
                Receive the inference format in the form of pd.DataFrame.
                ImageID and ClassProbability are essential for inferences.
        """
        super().__init__(data, inference)

        # check the existence of "ImageID" in data & inference
        if "ImageID" not in data:
            raise Exception("Invalid Data, ImageID not found in data")
        if "ImageID" not in inference:
            raise Exception("Invalid Data, ImageID not found in inference")

        # check the existence of "ClassProbability" in inference
        self.num_classes = 0
        for head in list(inference):
            if re.match(r"ClassProbability\d+", head):
                self.num_classes += 1

        if self.num_classes == 0:
            raise Exception("Invalid data, Inference do not have ClassProbability values")

        # rank: The inference DataFrame, sorted according to the score.
        self.rank = self._rank_images().sort_values(by="rank")

    def get_sample(self, method: str, k: int, n: int = 3) -> pd.DataFrame:
        """
        A function that extracts sample data and returns it.

        Args:
            method:
                - 'topk' - It extracts the k sample data with the
                  highest uncertainty.
                - 'lowk' - It extracts the k sample data with the
                  lowest uncertainty.
                - 'randomk' - Extract and return random k sample data.
            k: number of sample data
            n: Parameters to be used in the randtopk method, Variable to first
                extract data of multiple n of k.
        Returns:
            Extracted sample data : pd.DataFrame
        """
        temp_rank = self.rank

        # 1. k value check
        if not isinstance(k, int) or k <= 0:
            raise ValueError(f"Invalid value {k}. k must have an integer greater than zero.")

        # 2. Select a sample according to the method
        if k <= len(temp_rank):
            if method == self.sampling_method.topk.name:
                temp_rank = temp_rank[:k]
            elif method == self.sampling_method.lowk.name:
                temp_rank = temp_rank[-k:]
            elif method == self.sampling_method.randk.name:
                return self.data.sample(n=k).reset_index(drop=True)
            elif method in {self.sampling_method.mixk.name, self.sampling_method.randtopk.name}:
                return self._get_sample_mixed(method=method, k=k, n=n)
            else:
                raise ValueError(f"Unknown sampling method '{method}'")
        else:
            log.warning("The number of samples is greater than the size of the " "selected subset.")

        columns = list(self.data.columns)
        merged_df = pd.merge(temp_rank, self.data, how="inner", on=["ImageID"])
        return merged_df[columns].reset_index(drop=True)

    def _get_sample_mixed(self, method: str, k: int, n: int = 3) -> pd.DataFrame:
        """
        A function that extracts sample data and returns it.

        Args:
            method:
                - 'mixk': Return top-k and low-k halves based on uncertainty.
                - 'randomtopk': Randomly extract n*k and return k
                  with high uncertainty.
            k: number of sample data
            n: Number to extract n * k from total data according to n,
                and top-k from it
        Returns:
            Extracted sample data : pd.DataFrame
        """
        temp_rank = self.rank

        # Select a sample according to the method
        if k <= len(temp_rank):
            if method == self.sampling_method.mixk.name:
                if k % 2 == 0:
                    temp_rank = pd.concat([temp_rank[: k // 2], temp_rank[-(k // 2) :]])
                else:
                    temp_rank = pd.concat([temp_rank[: (k // 2) + 1], temp_rank[-(k // 2) :]])
            elif method == self.sampling_method.randtopk.name:
                if n * k <= len(temp_rank):
                    temp_rank = temp_rank.sample(n=n * k).sort_values(by="rank")
                else:
                    log.warning(msg="n * k exceeds the length of the inference")
                temp_rank = temp_rank[:k]

        columns = list(self.data.columns)
        merged_df = pd.merge(temp_rank, self.data, how="inner", on=["ImageID"])
        return merged_df[columns].reset_index(drop=True)

    def _rank_images(self) -> pd.DataFrame:
        """
        A internal function that ranks the inference data based on uncertainty.
        Returns:
            inference data sorted by uncertainty. pd.DataFrame
        """
        # 1. Load Inference
        inference, res = None, None
        if self.inference is not None:
            inference = pd.DataFrame(self.inference)
        else:
            raise Exception("Invalid Data, Failed to load inference result")

        # 2. If the reference data frame does not contain an uncertify score, calculate it
        if "Uncertainty" not in inference:
            inference = self._calculate_uncertainty_from_classprob(inference)

        # 3. Check that Uncertainty values are in place.
        na_df = inference.isna().sum()
        if "Uncertainty" in na_df and na_df["Uncertainty"] > 0:
            raise Exception("Some inference results do not have Uncertainty values")

        # 4. Ranked based on Uncertainty score
        res = inference[["ImageID", "Uncertainty"]].groupby("ImageID").mean()
        res["rank"] = res["Uncertainty"].rank(ascending=False, method="first")
        res = res.reset_index()

        return res

    def _calculate_uncertainty_from_classprob(self, inference: pd.DataFrame) -> pd.DataFrame:
        """
        A function that calculates uncertainty based on entropy through
        ClassProbability values.
        Args:
            inference: Inference data where uncertainty has not been calculated
        Returns:
            inference data with uncertainty variable
        """

        # Calculate Entropy (Uncertainty Score)
        uncertainty = []
        for i in range(len(inference)):
            entropy = 0
            for j in range(self.num_classes):
                p = inference.loc[i][f"ClassProbability{j+1}"]
                if p < 0 or p > 1:
                    raise Exception("Invalid data, Math domain Error! p is between 0 and 1")
                entropy -= p * math.log(p + 1e-14, math.e)

            uncertainty.append(entropy)

        inference["Uncertainty"] = uncertainty

        return inference



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from itertools import groupby
from typing import Callable, Dict, Iterable, NewType, Optional, Sequence, Tuple, Union

import numpy as np
from typing_extensions import Literal

from datumaro.components.annotation import AnnotationType, LabelCategories, Mask, RleMask, Shape
from datumaro.util.mask_tools import mask_to_rle


def find_instances(instance_anns):
    instance_anns = sorted(instance_anns, key=lambda a: a.group)
    ann_groups = []
    for g_id, group in groupby(instance_anns, lambda a: a.group):
        if not g_id:
            ann_groups.extend(([a] for a in group))
        else:
            ann_groups.append(list(group))

    return ann_groups


def find_group_leader(group):
    return max(group, key=lambda x: x.get_area())


BboxCoords = Tuple[float, float, float, float]
_Shape = NewType("_Shape", Shape)
SpatialAnnotation = Union[_Shape, Mask]


def _get_bbox(ann: Union[Sequence, SpatialAnnotation]) -> BboxCoords:
    if isinstance(ann, (Shape, Mask)):
        return ann.get_bbox()
    elif hasattr(ann, "__len__") and len(ann) == 4:
        return ann
    elif hasattr(ann, "__len__") and len(ann) == 0:
        return [0, 0, 0, 0]
    else:
        raise ValueError("The value of type '%s' can't be treated as a " "bounding box" % type(ann))


def max_bbox(annotations: Iterable[Union[BboxCoords, SpatialAnnotation]]) -> BboxCoords:
    """
    Computes the maximum bbox for the set of spatial annotations and boxes.

    Returns:
      bbox (tuple): (x, y, w, h)
    """

    boxes = [_get_bbox(ann) for ann in annotations]
    x0 = min((b[0] for b in boxes), default=0)
    y0 = min((b[1] for b in boxes), default=0)
    x1 = max((b[0] + b[2] for b in boxes), default=0)
    y1 = max((b[1] + b[3] for b in boxes), default=0)
    return [x0, y0, x1 - x0, y1 - y0]


def mean_bbox(annotations: Iterable[Union[BboxCoords, SpatialAnnotation]]) -> BboxCoords:
    """
    Computes the mean bbox for the set of spatial annotations and boxes.

    Returns:
      bbox (tuple): (x, y, w, h)
    """

    le = len(annotations)
    boxes = [_get_bbox(ann) for ann in annotations]
    mlb = sum(b[0] for b in boxes) / le
    mtb = sum(b[1] for b in boxes) / le
    mrb = sum(b[0] + b[2] for b in boxes) / le
    mbb = sum(b[1] + b[3] for b in boxes) / le
    return [mlb, mtb, mrb - mlb, mbb - mtb]


def softmax(x):
    return np.exp(x) / sum(np.exp(x))


def nms(segments, iou_thresh=0.5):
    """
    Non-maxima suppression algorithm.
    """

    indices = np.argsort([b.attributes["score"] for b in segments])
    ious = np.array([[segment_iou(a, b) for b in segments] for a in segments])

    predictions = []
    while len(indices) != 0:
        i = len(indices) - 1
        pred_idx = indices[i]
        to_remove = [i]
        predictions.append(segments[pred_idx])
        for i, box_idx in enumerate(indices[:i]):
            if iou_thresh < ious[pred_idx, box_idx]:
                to_remove.append(i)
        indices = np.delete(indices, to_remove)

    return predictions


def bbox_iou(a, b) -> Union[Literal[-1], float]:
    """
    IoU computations for simple cases with bounding boxes
    """
    bbox_a = _get_bbox(a)
    bbox_b = _get_bbox(b)

    aX, aY, aW, aH = bbox_a
    bX, bY, bW, bH = bbox_b
    in_right = min(aX + aW, bX + bW)
    in_left = max(aX, bX)
    in_top = max(aY, bY)
    in_bottom = min(aY + aH, bY + bH)

    in_w = max(0, in_right - in_left)
    in_h = max(0, in_bottom - in_top)
    intersection = in_w * in_h
    if not intersection:
        return -1

    a_area = aW * aH
    b_area = bW * bH
    union = a_area + b_area - intersection
    return intersection / union


def segment_iou(a, b):
    """
    Generic IoU computation with masks, polygons, and boxes.
    Returns -1 if no intersection, [0; 1] otherwise
    """
    from pycocotools import mask as mask_utils

    a_bbox = list(a.get_bbox())
    b_bbox = list(b.get_bbox())

    types = {a.type, b.type}
    is_bbox = AnnotationType.bbox in types or AnnotationType.ellipse in types
    if is_bbox:
        a = [a_bbox]
        b = [b_bbox]
    else:
        w = max(a_bbox[0] + a_bbox[2], b_bbox[0] + b_bbox[2])
        h = max(a_bbox[1] + a_bbox[3], b_bbox[1] + b_bbox[3])

        def _to_rle(ann):
            if ann.type == AnnotationType.polygon:
                return mask_utils.frPyObjects([ann.points], h, w)
            elif isinstance(ann, RleMask):
                return [ann.rle]
            elif ann.type == AnnotationType.mask:
                return mask_utils.frPyObjects([mask_to_rle(ann.image)], h, w)
            else:
                raise TypeError("Unexpected arguments: %s, %s" % (a, b))

        a = _to_rle(a)
        b = _to_rle(b)
    return float(mask_utils.iou(a, b, [not is_bbox]))


def PDJ(a, b, eps=None, ratio=0.05, bbox=None):
    """
    Percentage of Detected Joints metric.
    Counts the number of matching points.
    """

    assert eps is not None or ratio is not None

    p1 = np.array(a.points).reshape((-1, 2))
    p2 = np.array(b.points).reshape((-1, 2))
    if len(p1) != len(p2):
        return 0

    if not eps:
        if bbox is None:
            bbox = mean_bbox([a, b])

        diag = (bbox[2] ** 2 + bbox[3] ** 2) ** 0.5
        eps = ratio * diag

    dists = np.linalg.norm(p1 - p2, axis=1)
    return np.sum(dists < eps) / len(p1)


def OKS(a, b, sigma=None, bbox=None, scale=None):
    """
    Object Keypoint Similarity metric.
    https://cocodataset.org/#keypoints-eval
    """

    p1 = np.array(a.points).reshape((-1, 2))
    p2 = np.array(b.points).reshape((-1, 2))
    if len(p1) != len(p2):
        return 0

    if not sigma:
        sigma = 0.1
    else:
        assert len(sigma) == len(p1)

    if not scale:
        if bbox is None:
            bbox = mean_bbox([a, b])
        scale = bbox[2] * bbox[3]

    dists = np.linalg.norm(p1 - p2, axis=1)
    return np.sum(np.exp(-(dists**2) / (2 * scale * (2 * sigma) ** 2)))


def approximate_line(points: Sequence[float], segments: int) -> np.ndarray:
    """
    Approximates a 2d line to the required number of segments. The new points
    are distributed uniformly across the input line.

    Args:
      points (Sequence): an array of line point coordinates.
        The size is [points * 2], the layout is [x0, y0, x1, y1, ...].
      segments (int): the required numebr of segments in the resulting line.

    Returns:
        new_points (numpy.ndarray): an array of new line point coordinates.
    Returns:
        The size is [(segments + 1) * 2], the layout is [x0, y0, x1, y1, ...].
    """

    assert 2 <= len(points) // 2 and len(points) % 2 == 0
    assert 0 < segments

    points = list(points)
    if len(points) == 2:
        points.extend(points)
    points = np.array(points).reshape((-1, 2))

    lengths = np.linalg.norm(points[1:] - points[:-1], axis=1)
    dists = [0]
    for l in lengths:
        dists.append(dists[-1] + l)

    step = dists[-1] / segments

    new_points = np.zeros((segments + 1, 2))
    new_points[0] = points[0]

    old_segment = 0
    for new_segment in range(1, segments + 1):
        pos = new_segment * step
        while dists[old_segment + 1] < pos and old_segment + 2 < len(dists):
            old_segment += 1

        segment_start = dists[old_segment]
        segment_len = lengths[old_segment]
        prev_p = points[old_segment]
        next_p = points[old_segment + 1]
        r = (pos - segment_start) / segment_len

        new_points[new_segment] = prev_p * (1 - r) + next_p * r

    return np.reshape(new_points, (-1,))


def make_label_id_mapping(
    src_labels: LabelCategories, dst_labels: LabelCategories, fallback: int = 0
) -> Tuple[Callable[[int], Optional[int]], Dict[int, int], Dict[int, str], Dict[int, str]]:
    """
    Maps label ids from source to destination. Fallback id is used for missing
    labels.

    Returns:

    |   map_id (callable): src id -> dst id
    |   id_mapping (dict): src id -> dst i
    |   src_labels (dict): src id -> src label
    |   dst_labels (dict): dst id -> dst label
    """

    source_labels = {id: label.name for id, label in enumerate(src_labels or ())}
    target_labels = {label.name: id for id, label in enumerate(dst_labels or ())}
    id_mapping = {
        src_id: target_labels.get(src_label, fallback)
        for src_id, src_label in source_labels.items()
    }

    def map_id(src_id):
        return id_mapping.get(src_id, fallback)

    return map_id, id_mapping, source_labels, target_labels



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import OrderedDict
from enum import IntEnum


class SynthiaFormatType(IntEnum):
    synthia_rand = 0
    synthia_sf = 1
    synthia_al = 2


class SynthiaRandPath:
    IMAGES_DIR = "RGB"
    LABELS_SEGM_DIR = "GTTXT"
    SEMANTIC_SEGM_DIR = "GT"

    @classmethod
    def meta_folders(cls):
        return [val for val in vars(cls).values() if isinstance(val, str) and "datumaro" not in val]


class SynthiaSfPath:
    IMAGES_DIR = "RGBLeft"
    SEMANTIC_SEGM_DIR = "GTLeft"
    DEPTH_DIR = "DepthLeft"

    @classmethod
    def meta_folders(cls):
        return [val for val in vars(cls).values() if isinstance(val, str) and "datumaro" not in val]


class SynthiaAlPath:
    IMAGES_DIR = "RGB"
    SEMANTIC_SEGM_DIR = "SemSeg"
    DEPTH_DIR = "Depth"

    @classmethod
    def meta_folders(cls):
        return [val for val in vars(cls).values() if isinstance(val, str) and "datumaro" not in val]


SynthiaRandLabelMap = OrderedDict(
    [
        ("Void", (0, 0, 0)),
        ("Sky", (128, 128, 128)),
        ("Building", (128, 0, 0)),
        ("Road", (128, 64, 128)),
        ("Sidewalk", (0, 0, 192)),
        ("Fence", (64, 64, 128)),
        ("Vegetation", (128, 128, 0)),
        ("Pole", (192, 192, 128)),
        ("Car", (64, 0, 128)),
        ("Sign", (192, 128, 128)),
        ("Pedestrian", (64, 64, 0)),
        ("Cyclist", (0, 128, 192)),
        ("Lanemarking", (0, 175, 0)),
        ("TrafficLight", (0, 128, 128)),
    ]
)


SynthiaSfLabelMap = OrderedDict(
    [
        ("Void", (0, 0, 0)),
        ("Road", (128, 64, 128)),
        ("Sidewalk", (244, 35, 232)),
        ("Building", (70, 70, 70)),
        ("Wall", (102, 102, 156)),
        ("Fence", (190, 153, 153)),
        ("Pole", (153, 153, 153)),
        ("TrafficLight", (250, 170, 30)),
        ("TrafficSign", (220, 220, 0)),
        ("Vegetation", (107, 142, 35)),
        ("Terrian", (152, 251, 152)),
        ("Sky", (70, 130, 180)),
        ("Person", (220, 20, 60)),
        ("Rider", (255, 0, 0)),
        ("Car", (0, 0, 142)),
        ("Truck", (0, 80, 100)),
        ("Bus", (0, 60, 100)),
        ("Train", (0, 80, 100)),
        ("Motorcycle", (0, 0, 230)),
        ("Bicycle", (119, 11, 31)),
        ("RoadLines", (157, 234, 50)),
        ("Other", (72, 0, 98)),
        ("RoadWorks", (167, 106, 29)),
    ]
)


SynthiaAlLabelMap = OrderedDict(
    [
        ("Void", (0, 0, 0)),
        ("Sky", (128, 128, 128)),
        ("Building", (128, 0, 0)),
        ("Road", (128, 64, 128)),
        ("Sidewalk", (0, 0, 192)),
        ("Fence", (64, 64, 128)),
        ("Vegetation", (128, 128, 0)),
        ("Pole", (192, 192, 128)),
        ("Car", (64, 0, 128)),
        ("TrafficSign", (192, 128, 128)),
        ("Pedestrian", (64, 64, 0)),
        ("Bicycle", (0, 128, 192)),
        ("Lanemarking", (0, 172, 0)),
        ("TrafficLight", (0, 128, 128)),
    ]
)



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from enum import Enum, auto


class IcdarTask(Enum):
    word_recognition = auto()
    text_localization = auto()
    text_segmentation = auto()


class IcdarPath:
    IMAGE_EXT = ".png"
    GT_EXT = ".bmp"
    IMAGES_DIR = "images"



# Copyright (C) 2020 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import os.path as osp

from datumaro.components.annotation import AnnotationType, CompiledMask
from datumaro.components.errors import DatasetExportError, MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import Image
from datumaro.util.image import save_image
from datumaro.util.mask_tools import generate_colormap, paint_mask

from .format import IcdarPath


class IcdarWordRecognitionExporter(Exporter):
    DEFAULT_IMAGE_EXT = IcdarPath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        for subset_name, subset in self._extractor.subsets().items():
            annotation = ""
            for item in subset:
                image_filename = self._make_image_filename(item)
                if self._save_media and item.media:
                    self._save_image(
                        item,
                        osp.join(self._save_dir, subset_name, IcdarPath.IMAGES_DIR, image_filename),
                    )

                annotation += "%s, " % image_filename
                for ann in item.annotations:
                    if ann.type != AnnotationType.caption:
                        continue
                    annotation += '"%s"' % ann.caption
                annotation += "\n"

            if len(annotation):
                anno_file = osp.join(self._save_dir, subset_name, "gt.txt")
                os.makedirs(osp.dirname(anno_file), exist_ok=True)
                with open(anno_file, "w", encoding="utf-8") as f:
                    f.write(annotation)


class IcdarTextLocalizationExporter(Exporter):
    DEFAULT_IMAGE_EXT = IcdarPath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        for subset_name, subset in self._extractor.subsets().items():
            for item in subset:
                if self._save_media and item.media:
                    self._save_image(item, subdir=osp.join(subset_name, IcdarPath.IMAGES_DIR))

                annotation = ""
                for ann in item.annotations:
                    if ann.type == AnnotationType.bbox:
                        annotation += " ".join(str(p) for p in ann.points)
                        if ann.attributes and "text" in ann.attributes:
                            annotation += ' "%s"' % ann.attributes["text"]
                    elif ann.type == AnnotationType.polygon:
                        annotation += ",".join(str(p) for p in ann.points)
                        if ann.attributes and "text" in ann.attributes:
                            annotation += ',"%s"' % ann.attributes["text"]
                    annotation += "\n"

                anno_file = osp.join(
                    self._save_dir,
                    subset_name,
                    osp.dirname(item.id),
                    "gt_" + osp.basename(item.id) + ".txt",
                )
                os.makedirs(osp.dirname(anno_file), exist_ok=True)
                with open(anno_file, "w", encoding="utf-8") as f:
                    f.write(annotation)


class IcdarTextSegmentationExporter(Exporter):
    DEFAULT_IMAGE_EXT = IcdarPath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        for subset_name, subset in self._extractor.subsets().items():
            for item in subset:
                self._save_item(subset_name, subset, item)

    def _save_item(self, subset_name, subset, item):
        if self._save_media and item.media:
            self._save_image(item, subdir=osp.join(subset_name, IcdarPath.IMAGES_DIR))

        annotation = ""

        anns = [a for a in item.annotations if a.type == AnnotationType.mask]

        color_bank = iter(generate_colormap(len(anns), include_background=False).values())
        colormap = [(255, 255, 255)]
        used_colors = set(colormap)

        if anns:
            anns = sorted(anns, key=lambda a: int(a.attributes.get("index", 0)))
            group = anns[0].group
            for i, ann in enumerate(anns):
                # Assign new color if it is not defined
                color = ann.attributes.get("color", "")
                if color:
                    color = color.split()
                    if len(color) != 3:
                        raise DatasetExportError(
                            "Item %s: mask #%s has invalid color" % (item.id, i)
                        )

                    color = tuple(map(int, color))
                else:
                    color = next(color_bank)
                    while color in used_colors:
                        color = next(color_bank)
                colormap.append(color)
                used_colors.add(color)

                text = ann.attributes.get("text", "")
                bbox = ann.get_bbox()

                if ann.group != group or (not ann.group and anns[0].group != 0):
                    annotation += "\n"
                if text == " ":
                    annotation += "#"
                annotation += " ".join(str(p) for p in color)
                annotation += " %s" % ann.attributes.get("center", "- -")
                annotation += " %s %s %s %s" % (
                    bbox[0],
                    bbox[1],
                    bbox[0] + bbox[2],
                    bbox[1] + bbox[3],
                )
                annotation += ' "%s"' % text
                annotation += "\n"
                group = ann.group

            mask = CompiledMask.from_instance_masks(
                anns, instance_labels=[m.attributes["index"] + 1 for m in anns]
            )
            mask = paint_mask(mask.class_mask, {i: colormap[i] for i in range(len(colormap))})
            save_image(
                osp.join(self._save_dir, subset_name, item.id + "_GT" + IcdarPath.GT_EXT),
                mask,
                create_dir=True,
            )

        anno_file = osp.join(self._save_dir, subset_name, item.id + "_GT" + ".txt")
        os.makedirs(osp.dirname(anno_file), exist_ok=True)
        with open(anno_file, "w", encoding="utf-8") as f:
            f.write(annotation)



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import warnings

warnings.warn(
    "We are planning to clean up the entities in datumaro/__init__.py until datumaro==1.5.0. "
    'This will affect the following import pattern: "import datumaro as dm; dm.<entity>". '
    "If you are using this pattern in your code, please revisit it after upgrading datumaro>=1.5.0.",
    DeprecationWarning,
)

from . import errors as errors
from . import ops as ops
from . import project as project
from .components.algorithms import RISE, LossDynamicsAnalyzer
from .components.annotation import (
    NO_GROUP,
    Annotation,
    AnnotationType,
    Bbox,
    BinaryMaskImage,
    Caption,
    Categories,
    Colormap,
    CompiledMask,
    CompiledMaskImage,
    Cuboid3d,
    Ellipse,
    IndexMaskImage,
    Label,
    LabelCategories,
    Mask,
    MaskCategories,
    Points,
    PointsCategories,
    Polygon,
    PolyLine,
    RgbColor,
    RleMask,
)
from .components.cli_plugin import CliPlugin
from .components.dataset import Dataset, DatasetPatch, DatasetSubset, StreamDataset, eager_mode
from .components.dataset_base import (
    DEFAULT_SUBSET_NAME,
    CategoriesInfo,
    DatasetBase,
    DatasetItem,
    IDataset,
    SubsetBase,
)
from .components.dataset_item_storage import ItemStatus
from .components.environment import Environment
from .components.exporter import Exporter, ExportErrorPolicy, FailingExportErrorPolicy
from .components.hl_ops import HLOps
from .components.importer import FailingImportErrorPolicy, Importer, ImportErrorPolicy
from .components.launcher import Launcher
from .components.media import Image, MediaElement, Video, VideoFrame
from .components.progress_reporting import (
    NullProgressReporter,
    ProgressReporter,
    SimpleProgressReporter,
    TQDMProgressReporter,
)
from .components.registry import PluginRegistry
from .components.transformer import ItemTransform, ModelTransform, Transform
from .components.validator import Validator
from .components.visualizer import Visualizer
from .version import __version__



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp
from enum import Enum

from datumaro.components.environment import Environment
from datumaro.components.errors import MigrationError, ProjectNotFoundError
from datumaro.components.filter import DatasetItemEncoder
from datumaro.components.operations import compute_ann_statistics, compute_image_statistics
from datumaro.components.project import Project, ProjectBuildTargets
from datumaro.components.validator import TaskType
from datumaro.util import dump_json_file, str_to_bool
from datumaro.util.os_util import make_file_name
from datumaro.util.scope import scope_add, scoped

from ...commands import get_project_commands
from ...util import MultilineFormatter, add_subparser, make_subcommands_help
from ...util.errors import CliException
from ...util.project import generate_next_file_name, load_project, parse_full_revpath


def build_migrate_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Migrate project",
        description="""
        Migrates the project from the old version to a new one.|n
        |n
        Examples:|n
        - Migrate a project from v1 to v2, save the new project in other dir:|n
        |s|s%(prog)s -o <output/dir>
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        required=True,
        help="Output directory for the updated project",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Ignore source import errors (default: %(default)s)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to migrate (default: current dir)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.set_defaults(command=migrate_command)

    return parser


def get_migrate_sensitive_args():
    return {
        migrate_command: [
            "dst_dir",
            "project_dir",
        ],
    }


@scoped
def migrate_command(args):
    dst_dir = args.dst_dir
    if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
        raise CliException(
            "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
        )
    dst_dir = osp.abspath(dst_dir)

    log.debug("Migrating project from v1 to v2...")

    try:
        Project.migrate_from_v1_to_v2(args.project_dir, dst_dir, skip_import_errors=args.force)
    except Exception as e:
        raise MigrationError(
            "Failed to migrate the project "
            "automatically. Try to create a new project and import sources "
            "manually with 'datum create' and 'datum import'."
        ) from e

    log.debug("Finished")


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        description="""
            Manipulate projects.|n
            |n
            By default, the project to be operated on is searched for
            in the current directory. An additional '-p' argument can be
            passed to specify project location.
        """,
        formatter_class=MultilineFormatter,
    )

    subcommands_desc = ""
    known_commands = get_project_commands()
    help_line_start = max((len(e[0]) for e in known_commands), default=0)
    help_line_start = max((2 + help_line_start) // 4 + 1, 6) * 4  # align to tabs
    if known_commands:
        if subcommands_desc:
            subcommands_desc += "\n"
        subcommands_desc += make_subcommands_help(known_commands, help_line_start)
    if subcommands_desc:
        subcommands_desc += (
            "\nRun '%s COMMAND --help' for more information on a command." % parser.prog
        )

    # TODO: revisit project related commands for better code structure
    subparsers = parser.add_subparsers(
        title=subcommands_desc, description="", help=argparse.SUPPRESS
    )

    commands = get_project_commands()
    for command_name, command, _ in commands:
        if command is not None:
            add_subparser(subparsers, command_name, command.build_parser)

    add_subparser(subparsers, "migrate", build_migrate_parser)

    return parser


def get_sensitive_args():
    return {
        **get_migrate_sensitive_args(),
    }



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT


import argparse
import logging as log

from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.errors import ProjectNotFoundError
from datumaro.util.scope import on_error_do, scope_add, scoped

from ....util import MultilineFormatter, join_cli_args, show_video_import_warning
from ....util.errors import CliException
from ....util.project import load_project

__all__ = [
    "build_parser",
    "get_sensitive_args",
]


def build_parser(parser_ctor=argparse.ArgumentParser):
    env = DEFAULT_ENVIRONMENT
    builtins = sorted(set(env.extractors) | set(env.importers))

    parser = parser_ctor(
        help="Add data source to project",
        description="""
        Adds a data source to a project. A data source is a dataset
        in a supported format (check 'formats' section below).|n
        |n
        The command adds a project-local directory as a data source in
        the project. Unlike the "import" command, it does not copy datasets
        and only works with local directories. The source name is defined by
        the directory name.|n
        |n
        Formats:|n
        Datasets come in a wide variety of formats. Each dataset
        format defines its own data structure and rules on how to
        interpret the data. Check the user manual for the list of
        supported formats, examples and documentation.
        |n
        The list of supported formats can be extended by plugins.
        Check the "plugins" section of the developer guide for information
        about plugin implementation.|n
        |n
        Each dataset format has its own import options, which are passed
        after the '--' separator (see examples), pass '-- -h' for more info.|n
        |n
        Builtin formats: {}|n
        |n
        Examples:|n
        - Add a local directory with a VOC-like dataset:|n
        |s|s%(prog)s -f voc my_dataset/|n
        |n
        - Add a directory with a COCO dataset, use only a specific file:|n
        |s|s%(prog)s -f coco_instances -r anns/train.json my_source/
        """.format(
            ", ".join(builtins)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument("path", help="A path to a dataset directory")
    parser.add_argument("-f", "--format", required=True, help="Source dataset format")
    parser.add_argument(
        "-r",
        "--path",
        dest="rpath",
        help="A path relative to source root to the source data. Useful to "
        "specify a path to subset, subtask, or a specific file in dataset.",
    )
    parser.add_argument("--no-check", action="store_true", help="Don't try to read the source")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        default=".",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for extractor (pass '-- -h' for help). "
        "Must be specified after the main command arguments and after "
        "the '--' separator",
    )
    parser.set_defaults(command=add_command)

    return parser


def get_sensitive_args():
    return {
        add_command: ["url", "project_dir", "rpath", "extra_args"],
    }


@scoped
def add_command(args):
    # Workaround. Required positionals consume positionals from the end
    args._positionals += join_cli_args(args, "path", "extra_args")

    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
    else:
        pos = 1
    args.path = (args._positionals[:pos] or [""])[0]
    args.extra_args = args._positionals[pos + has_sep :]

    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help:
            raise

    if project is not None:
        env = project.env
    else:
        env = DEFAULT_ENVIRONMENT

    fmt = args.format
    if fmt in env.importers:
        arg_parser = env.importers[fmt]
    elif fmt in env.extractors:
        arg_parser = env.extractors[fmt]
    else:
        raise CliException(
            "Unknown format '%s'. A format can be added"
            " by providing an Extractor and Importer plugins" % fmt
        )

    extra_args = arg_parser.parse_cmdline(args.extra_args)

    if fmt == "video_frames":
        show_video_import_warning()

    name, _ = project.add_source(
        args.path, format=args.format, options=extra_args, rpath=args.rpath
    )
    on_error_do(
        project.remove_source, name, ignore_errors=True, kwargs={"force": True, "keep_data": True}
    )

    if not args.no_check:
        log.info("Checking the source...")
        project.working_tree.make_dataset(name)

    project.working_tree.save()

    log.info("Source '%s' with format '%s' has been added to the project", name, args.format)

    return 0



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.samples.utils import gen_hash_key


class ClipTextViTB32ModelInterpreter(IModelInterpreter):
    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        return img, None

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        feature_vector = pred.get("output")
        if feature_vector is None:
            raise DatumaroError('"output" key should exist in the model prediction.')

        return [gen_hash_key(feature_vector)]

    def get_categories(self):
        label_categories = LabelCategories()
        return {AnnotationType.label: label_categories}



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import csv
import errno
import os
import os.path as osp
from typing import List, Optional

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Video
from datumaro.plugins.data_formats.video import VIDEO_EXTENSIONS
from datumaro.rust_api import JsonSectionPageMapper
from datumaro.util import parse_json, parse_json_file
from datumaro.util.os_util import find_files


class KineticsBase(DatasetBase):
    def __init__(self, path: str, *, ctx: Optional[ImportContext] = None):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)
        self._path = path

        super().__init__(media_type=Video, ctx=ctx)

        self._annotation_files = {}
        for ann_file in find_files(path, ["csv", "json"]):
            filename = osp.splitext(osp.basename(ann_file))[0]
            if (filename not in self._annotation_files) or (
                self._annotation_files.get(filename, "").endswith("json")
            ):
                self._annotation_files[filename] = ann_file

        self._subsets = {}
        self._subset_media_files = {}
        self._categories = {AnnotationType.label: LabelCategories()}
        self._items = []

        for ann_file in sorted(self._annotation_files.values()):
            if ann_file.endswith("csv"):
                self._load_items_from_csv(ann_file)
            else:
                self._load_items_from_json(ann_file)

    def __iter__(self):
        return iter(self._items)

    def categories(self):
        return self._categories

    def _subset_path(self, subset):
        if subset in self._subsets:
            return self._subsets[subset]

        subset_path = osp.join(self._path, subset)
        self._subsets[subset] = subset_path if osp.isdir(subset_path) else self._path

        return self._subsets[subset]

    def _media_files(self, subset):
        if subset in self._subset_media_files:
            return self._subset_media_files[subset]

        subset_path = self._subset_path(subset)

        self._subset_media_files[subset] = {}
        for root, _, files in os.walk(subset_path):
            for f in files:
                key, file_extension = osp.splitext(f)
                if file_extension.lstrip(".") in VIDEO_EXTENSIONS:
                    self._subset_media_files[subset][key] = osp.join(root, f)

        return self._subset_media_files[subset]

    def _search_media_by_id(self, media_files, item_id):
        if item_id in media_files:
            return media_files[item_id]

        for media_file, file_path in media_files.items():
            if -1 < media_file.rfind(item_id):
                return file_path

        return None

    def _create_item(self, item_id, label, time_start, time_end, subset):
        label_id, _ = self._categories[AnnotationType.label].find(label)
        if label_id is None:
            label_id = self._categories[AnnotationType.label].add(label)

        media_path = self._search_media_by_id(self._media_files(subset), item_id)
        media = Video(media_path) if media_path else None

        self._ann_types.add(AnnotationType.label)

        return DatasetItem(
            id=item_id,
            annotations=[
                Label(
                    label=label_id,
                    attributes={"time_start": int(time_start), "time_end": int(time_end)},
                )
            ],
            subset=subset,
            media=media,
        )

    def _load_items_from_csv(self, path):
        with open(path, "r", encoding="utf-8") as f:
            for item_desc in csv.DictReader(f):
                self._items.append(
                    self._create_item(
                        item_desc["youtube_id"],
                        item_desc["label"],
                        item_desc["time_start"],
                        item_desc["time_end"],
                        item_desc["split"],
                    )
                )

    def _load_items_from_json(self, path):
        subset_desc = parse_json_file(path)
        for item_id, item_desc in subset_desc.items():
            self._items.append(
                self._create_item(
                    item_id,
                    item_desc["annotations"]["label"],
                    item_desc["annotations"]["segment"][0],
                    item_desc["annotations"]["segment"][1],
                    item_desc["subset"],
                )
            )


class KineticsImporter(Importer):
    _ANNO_EXTS = [".json", ".csv"]

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        with context.require_any():
            with context.alternative():
                ann_file = context.require_file(f"*{cls._ANNO_EXTS[0]}")

                with context.probe_text_file(
                    ann_file,
                    "JSON file must contain an youtube 'url' key",
                ) as f:
                    fpath = osp.join(context.root_path, ann_file)
                    page_mapper = JsonSectionPageMapper(fpath)
                    sections = page_mapper.sections()

                    page_map = next(iter(sections.values()))
                    offset, size = page_map["offset"], page_map["size"]

                    f.seek(offset, 0)
                    contents = parse_json(f.read(size))
                    if not isinstance(contents, dict):
                        raise Exception
                    if "youtube" not in contents.get("url", ""):
                        raise Exception

            with context.alternative():
                ann_file = context.require_file(f"*{cls._ANNO_EXTS[1]}")

                with context.probe_text_file(
                    ann_file,
                    "CSV file must contain 'youtube_id' in header",
                ) as f:
                    header = f.readline()
                    if "youtube_id" not in header.split(","):
                        raise Exception

    @classmethod
    def find_sources(cls, path):
        if find_files(path, ["csv", "json"]):
            return [{"url": path, "format": KineticsBase.NAME}]
        return []

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return cls._ANNO_EXTS



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from enum import Enum, auto


class CocoTask(Enum):
    instances = auto()
    person_keypoints = auto()
    captions = auto()
    labels = auto()  # extension, does not exist in the original COCO format
    image_info = auto()
    panoptic = auto()
    stuff = auto()


class CocoImporterType(Enum):
    default = auto()
    roboflow = auto()
    mmdet = auto()


class CocoPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"

    IMAGE_EXT = ".jpg"
    PANOPTIC_EXT = ".png"



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
import glob
import logging as log
import os.path as osp
from typing import List

from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME
from datumaro.components.errors import DatasetNotFoundError
from datumaro.components.importer import Importer
from datumaro.util import str_to_bool

from .base import MapillaryVistasInstancesBase, MapillaryVistasPanopticBase
from .format import MapillaryVistasPath, MapillaryVistasTask


class MapillaryVistasImporter(Importer):
    _TASKS = {
        MapillaryVistasTask.instances: MapillaryVistasInstancesBase,
        MapillaryVistasTask.panoptic: MapillaryVistasPanopticBase,
    }

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--format-version",
            default="v2.0",
            type=str,
            help="Use original config*.json file for your version of dataset",
        )
        parser.add_argument(
            "--parse-polygon",
            type=str_to_bool,
            default=False,
            help="Use original config*.json file for your version of dataset",
        )
        parser.add_argument(
            "--use-original-config",
            action="store_true",
            help="Use original config*.json file for your version of dataset",
        )
        parser.add_argument(
            "--keep-original-category-ids",
            action="store_true",
            help="Add dummy label categories so that category indices "
            "correspond to the category IDs in the original annotation "
            "file",
        )
        return parser

    def __call__(self, path, **extra_params):
        subsets = self.find_sources(path)

        if len(subsets) == 0:
            raise DatasetNotFoundError(path, self.NAME)

        tasks = list(set(task for subset in subsets.values() for task in subset))
        selected_task = tasks[0]
        if 1 < len(tasks):
            task_types = ",".join(task.name for task in tasks)
            log.warning(
                f"Found potentially conflicting source types: {task_types}"
                f"Only one one type will be used: {selected_task.name}"
            )

        if selected_task == MapillaryVistasTask.instances:
            has_config = any(
                [
                    osp.isfile(osp.join(path, config))
                    for config in MapillaryVistasPath.CONFIG_FILES.values()
                ]
            )

            if not has_config and not extra_params.get("use_original_config"):
                raise DatasetNotFoundError(
                    path,
                    self.NAME,
                    "Failed to find config*.json at '{path}'. "
                    "See extra args for using original configs.",
                )

        sources = [
            {"url": url, "format": self._TASKS[task].NAME, "options": dict(extra_params)}
            for _, subset_info in subsets.items()
            for task, url in subset_info.items()
            if task == selected_task
        ]

        return sources

    @classmethod
    def find_sources(cls, path):
        subsets = {}

        suffixes = [
            osp.join(ann_dir, subdir)
            for ann_dir, subdirs in MapillaryVistasPath.ANNOTATION_DIRS.items()
            for subdir in subdirs
        ]

        for suffix in suffixes:
            task = MapillaryVistasPath.CLASS_BY_DIR[osp.basename(suffix)]
            if task not in cls._TASKS:
                continue

            if osp.isdir(osp.join(path, suffix)):
                return {DEFAULT_SUBSET_NAME: {task: path}}

            for ann_path in glob.glob(osp.join(path, "*", suffix)):
                subset = osp.dirname(osp.dirname(osp.relpath(ann_path, path)))
                subsets.setdefault(subset, {})[task] = osp.join(path, subset)

        return subsets

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [".jpg", ".png", ".json"]


class MapillaryVistasInstancesImporter(MapillaryVistasImporter):
    _TASK = MapillaryVistasTask.instances
    _TASKS = {_TASK: MapillaryVistasImporter._TASKS[_TASK]}


class MapillaryVistasPanopticImporter(MapillaryVistasImporter):
    _TASK = MapillaryVistasTask.panoptic
    _TASKS = {_TASK: MapillaryVistasImporter._TASKS[_TASK]}
